# type: ignore
"""Multiple comparison corrections."""
import copy
import math
from typing import List, Optional, Tuple, Union

import numpy as np
from numpy import concatenate as cat
from scipy.interpolate import interp1d
from scipy.linalg import toeplitz
from scipy.sparse import csr_matrix
from scipy.special import betaln, gamma, gammaln

from brainstat._typing import ArrayLike
from brainstat.mesh.utils import mesh_edges
from brainstat.stats.utils import colon, interp1, ismember, row_ismember


def _fdr(self) -> np.ndarray:
    """Q-values for False Discovey Rate of resels.

    Parameters
    ----------
    self : brainstat.stats.SLM.SLM
        SLM object with computed t-values.

    Returns
    -------
    numpy.array
        Q-values for false discovery rate of resels.

    """
    l, v = np.shape(self.t)

    if self.mask is None:
        self.mask = np.ones((v), dtype="bool")

    df = np.zeros((2, 2))
    ndf = len(np.array([self.df]))
    df[0, 0:ndf] = self.df
    df[1, 0:2] = np.array([self.df])[ndf - 1]

    if self.dfs is not None:
        df[0, ndf - 1] = self.dfs[0, self.mask > 0].mean()

    if self.du is not None:
        resels, reselspvert, edg = compute_resels(self)
    else:
        reselspvert = np.ones((v))

    varA = np.append(10, self.t[0, self.mask.astype(bool)])
    P_val = stat_threshold(df=df, p_val_peak=varA, nvar=self.k, nprint=0)[0]
    P_val = P_val[1 : len(P_val)]
    nx = len(P_val)
    index = P_val.argsort()
    P_sort = P_val[index]
    r_sort = reselspvert[index]
    c_sort = np.cumsum(r_sort)
    P_sort = P_sort / (c_sort + (c_sort <= 0)) * (c_sort > 0) * r_sort.sum()
    m = 1
    Q_sort = np.zeros((1, nx))

    for i in np.arange(nx, 0, -1):
        if P_sort[i - 1] < m:
            m = P_sort[i - 1]
        Q_sort[0, i - 1] = m

    Q_tmp = np.zeros((1, nx))
    Q_tmp[0, index] = Q_sort

    Q = np.ones((self.mask.shape[0]))
    Q[self.mask] = np.squeeze(Q_tmp[0, :])

    return Q


valid_rft_output = Union[
    Tuple[dict, dict, dict, np.ndarray], Tuple[dict, List, List, List]
]


def _random_field_theory(self) -> valid_rft_output:
    """Corrected P-values for vertices and clusters.
    Parameters
    ----------
    self : brainstat.stats.SLM.SLM
        SLM object with computed t-values.

    Returns
    -------
    pval : a dictionary with keys 'P', 'C', 'mask'.
        pval['P'] : 2D numpy array of shape (1,v).
            Corrected P-values for vertices.
        pval['C'] : 2D numpy array of shape (1,v).
            Corrected P-values for clusters.
    peak : a dictionary with keys 't', 'vertid', 'clusid', 'P'.
        peak['t'] : 2D numpy array of shape (np,1).
            Peaks (local maxima).
        peak['vertid'] : 2D numpy array of shape (np,1).
            Vertex.
        peak['clusid'] : 2D numpy array of shape (np,1).
            Cluster id numbers.
        peak['P'] : 2D numpy array of shape (np,1).
            Corrected P-values for the peak.
    clus : a dictionary with keys 'clusid', 'nverts', 'resels', 'P.'
        clus['clusid'] : 2D numpy array of shape (nc,1).
            Cluster id numbers
        clus['nverts'] : 2D numpy array of shape (nc,1).
            Number of vertices in cluster.
        clus['resels'] : 2D numpy array of shape (nc,1).
            resels in the cluster.
        clus['P'] : 2D numpy array of shape (nc,1).
            Corrected P-values for the cluster.
    clusid : 2D numpy array of shape (1,v).
        Cluster id's for each vertex.
    Reference: Worsley, K.J., Andermann, M., Koulis, T., MacDonald, D.
    & Evans, A.C. (1999). Detecting changes in nonisotropic images.
    Human Brain Mapping, 8:98-101.
    """
    l, v = np.shape(self.t)

    if self.mask is None:
        self.mask = np.ones((v), dtype=bool)

    df = np.zeros((2, 2))
    ndf = len(np.array([self.df]))
    df[0, 0:ndf] = self.df
    df[1, 0:2] = np.array([self.df])[ndf - 1]

    if self.dfs is not None:
        df[0, ndf - 1] = self.dfs[0, self.mask > 0].mean()

    if v == 1:
        varA = varA = np.concatenate((np.array([10]), self.t[0]))
        pval = {}
        pval["P"] = stat_threshold(df=df, p_val_peak=varA, nvar=self.k, nprint=0)[0]
        pval["P"] = pval["P"][1]
        peak = []
        clus = []
        clusid = []
        # only a single p-value is returned, and function is stopped.
        return pval, peak, clus, clusid

    if self.cluster_threshold < 1:
        thresh_tmp = stat_threshold(
            df=df, p_val_peak=self.cluster_threshold, nvar=self.k, nprint=0
        )[0]
        thresh = float(thresh_tmp[0])
    else:
        thresh = self.cluster_threshold

    resels, reselspvert, edg = compute_resels(self)
    N = self.mask.sum()

    if np.max(self.t[0, self.mask]) < thresh:
        pval = {}
        varA = np.concatenate((np.array([[10]]), self.t), axis=1)
        pval["P"] = stat_threshold(
            search_volume=resels,
            num_voxels=N,
            fwhm=1,
            df=df,
            p_val_peak=varA.flatten(),
            nvar=self.k,
            nprint=0,
        )[0]
        pval["P"] = pval["P"][1 : v + 1]
        pval["C"] = None
        peak = {"t": None, "clusid": None, "vertid": None, "P": None}
        clus = {"clusid": None, "nverts": None, "resels": None, "P": None}
        clusid = None
    else:
        peak, clus, clusid = peak_clus(self, thresh, reselspvert, edg)
        self.t = self.t.reshape(1, self.t.size)
        varA = np.concatenate((np.array([[10]]), peak["t"].T, self.t), axis=1)
        varB = np.concatenate((np.array([[10]]), clus["resels"]))
        pp, clpval, _, _, _, _, = stat_threshold(
            search_volume=resels,
            num_voxels=N,
            fwhm=1,
            df=df,
            p_val_peak=varA.flatten(),
            cluster_threshold=thresh,
            p_val_extent=varB,
            nvar=float(self.k),
            nprint=0,
        )
        lenPP = len(pp[1 : len(peak["t"]) + 1])
        peak["P"] = pp[1 : len(peak["t"]) + 1].reshape(lenPP, 1)
        pval = {}
        pval["P"] = pp[len(peak["t"]) + np.arange(1, v + 1)]

        if self.k > 1:
            j = np.arange(self.k)[::-2]
            sphere = np.zeros((1, int(self.k)))
            sphere[:, j] = np.exp(
                (j + 1) * np.log(2)
                + (j / 2) * np.log(math.pi)
                + gammaln((self.k + 1) / 2)
                - gammaln(j + 1)
                - gammaln((self.k + 1 - j) / 2)
            )
            sphere = sphere * np.power(4 * np.log(2), -np.arange(0, self.k) / 2) / ndf
            varA = np.convolve(resels.flatten(), sphere.flatten())
            varB = np.concatenate((np.array([[10]]), clus["resels"]))
            pp, clpval, _, _, _, _, = stat_threshold(
                search_volume=varA,
                num_voxels=math.inf,
                fwhm=1.0,
                df=df,
                cluster_threshold=thresh,
                p_val_extent=varB,
                nprint=0,
            )

        clus["P"] = clpval[1 : len(clpval)]
        x = np.concatenate((np.array([[0]]), clus["clusid"]), axis=0)
        y = np.concatenate((np.array([[1]]), clus["P"]), axis=0)
        pval["C"] = interp1d(x.flatten(), y.flatten())(clusid)

    tlim = stat_threshold(
        search_volume=resels,
        num_voxels=N,
        fwhm=1,
        df=df,
        p_val_peak=np.array([0.5, 1]),
        nvar=float(self.k),
        nprint=0,
    )[0]
    tlim = tlim[1]
    pval["P"] = pval["P"] * (self.t[0, :] > tlim) + (self.t[0, :] <= tlim)

    return pval, peak, clus, clusid


def stat_threshold(
    search_volume: Union[float, ArrayLike] = 0,
    num_voxels: Union[float, ArrayLike] = 1,
    fwhm: float = 0.0,
    df: Union[float, ArrayLike] = math.inf,
    p_val_peak: float = 0.05,
    cluster_threshold: float = 0.001,
    p_val_extent: Union[float, ArrayLike] = 0.05,
    nconj: float = 1,
    nvar: int = 1,
    EC_file: Optional[bool] = None,
    nprint: int = 5,
) -> Tuple[np.ndarray, ...]:
    """Thresholds and P-values of peaks and clusters of random fields in any D.

    Parameters
    ----------
    search_volume : a float, or a list, or a numpy array
        volume of the search region in mm^3.
    num_voxels : a float, or int, or list, or 1D numpy array
        number of voxels (3D) or pixels (2D) in the search volume.
    fwhm : a float, or int.
        fwhm in mm of a smoothing kernel applied to the data.
    df : a float, or int, or list, or  array of shape (2,2)
        degrees of freedom.
    p_val_peak : a float, or 1D array of shape (y,)
        desired P-values for peaks.
    cluster_threshold: a float
        scalar threshold of the image for clusters
    p_val_extent : a float, or list, or 1D array of shape (y,)
        desired P-values for spatial extents of clusters of contiguous
        voxels above the cluster_threshold
    nconj : a float, or int
        number of conjunctions
    nvar :an int, list or 1D array of 1 or 2 integers
        number of variables for multivariate equivalents of T and F
        statistics

    Returns
    -------
    peak_threshold :
        thresholds for local maxima or peaks
    extent_threshold :

    peak_threshold_1
        height of a single peak chosen in advance
    extent_threshold_1
        extent of a single cluster chosen in advance
    t :

    rho :
    """

    def gammalni(n):
        x = math.inf * np.ones(n.shape)
        x[n >= 0] = gammaln(n[n >= 0])
        return x

    def minterp1(x, y, ix):
        # interpolates only the monotonically increasing values of x at ix
        n = x.size
        ix = np.array(ix)
        ix_shape = ix.shape
        ix = ix.flatten("F")

        mx = np.array(x[0], ndmin=1)
        my = np.array(y[0], ndmin=1)
        xx = x[0]
        for i in range(0, n):
            if x[i] > xx:
                xx = x[i]
                mx = np.append(mx, xx)
                my = np.append(my, y[i])

        out = []
        for i in range(0, ix.size):
            if ix[i] < mx[0] or ix[i] > mx[-1]:
                out.append(math.nan)
            else:
                out.append(interp1(mx, my, ix[i]))
        out = np.reshape(out, ix_shape, order="F")
        return out

    # Deal with the input

    # Make sure all input is in np.array format.
    fwhm = np.array(fwhm, ndmin=1)
    search_volume = np.array(search_volume, ndmin=2)
    num_voxels = np.array(num_voxels)
    df = np.array(df, ndmin=2)
    nvar = np.array(nvar, dtype=int)
    p_val_peak = np.array(p_val_peak, ndmin=1)
    p_val_extent = np.array(p_val_extent, ndmin=1)

    # Set the FWHM
    if fwhm.ndim == 1:
        fwhm = np.expand_dims(fwhm, axis=0)
        fwhm = np.r_[fwhm, fwhm]
    if np.shape(fwhm)[1] == 1:
        scale = 1
    else:
        scale = fwhm[0, 1] / fwhm[0, 0]
        fwhm = fwhm[:, 0]
    isscale = scale > 1

    # Set the number of voxels
    if num_voxels.size == 1:
        num_voxels = np.append(num_voxels, 1)

    # Set the search volume.
    if search_volume.shape[1] == 1:
        radius = (search_volume / (4 / 3 * math.pi)) ** (1 / 3)
        search_volume = np.c_[
            np.ones(radius.shape), 4 * radius, 2 * radius ** 2 * math.pi, search_volume
        ]

    if search_volume.shape[0] == 1:
        search_volume = np.concatenate(
            (
                search_volume,
                np.concatenate(
                    (np.ones((1, 1)), np.zeros((1, search_volume.size - 1))), axis=1
                ),
            ),
            axis=0,
        )

    lsv = search_volume.shape[1]
    if all(fwhm > 0):
        fwhm_inv = all(fwhm > 0) / fwhm + any(fwhm <= 0)
    else:
        fwhm_inv = np.zeros(fwhm.shape)
    if fwhm_inv.ndim == 1:
        fwhm_inv = np.expand_dims(fwhm_inv, axis=1)

    resels = search_volume * fwhm_inv ** np.arange(0, lsv)
    invol = resels * (4 * np.log(2)) ** (np.arange(0, lsv) / 2)

    D = []
    for k in range(2):
        D.append(np.max(np.argwhere(invol[k, :])))

    # determines which method was used to estimate fwhm (see fmrilm or multistat):
    df_limit = 4

    if df.size == 1:
        df = np.c_[df, np.zeros((1, 1))]
    if df.shape[0] == 1:
        infs = np.array([math.inf, math.inf], ndmin=2)
        df = np.r_[df, infs, infs]
    if df.shape[1] == 1:
        df = np.c_[df, df]
        df[0, 1] = 0
    if df.shape[0] == 2:
        df = np.r_[df, np.expand_dims(df[1, :], axis=0)]

    # is_tstat=1 if it is a t statistic
    is_tstat = df[0, 1] == 0
    if is_tstat:
        df1 = 1
        df2 = df[0, 0]
    else:
        df1 = df[0, 0]
        df2 = df[0, 1]
    if df2 >= 1000:
        df2 = math.inf
    df0 = df1 + df2

    dfw1 = df[1:3, 0].astype("float64")
    dfw2 = df[1:3, 1].astype("float64")
    dfw1[dfw1 >= 1000] = math.inf
    dfw2[dfw2 >= 1000] = math.inf
    if nvar.size == 1:
        nvar = np.r_[nvar, int(np.round(df1))]

    if isscale and (D[1] > 1 or nvar[0] > 1 | df2 < math.inf):
        print(D)
        print(nvar)
        print(df2)
        print("Cannot do scale space.")
        return
    Dlim = D + np.array([scale > 1, 0])
    DD = Dlim + nvar - 1

    # Values of the F statistic:
    t = (np.arange(1000, 0, -1) / 100) ** 4

    # Find the upper tail probs cumulating the F density using Simpson's rule:
    if math.isinf(df2):
        u = df1 * t
        b = (
            np.exp(-u / 2 - np.log(2 * math.pi) / 2 + np.log(u) / 4)
            * df1 ** (1 / 4)
            * 4
            / 100
        )
    else:
        u = df1 * t / df2
        b = (
            np.exp(
                -df0 / 2 * np.log(1 + u) + np.log(u) / 4 - betaln(1 / 2, (df0 - 1) / 2)
            )
            * (df1 / df2) ** (1 / 4)
            * 4
            / 100
        )

    t = np.r_[t, 0]
    b = np.r_[b, 0]
    n = t.size
    sb = np.cumsum(b)
    sb1 = np.cumsum(b * (-1) ** np.arange(1, n + 1))
    pt1 = sb + sb1 / 3 - b / 3
    pt2 = sb - sb1 / 3 - b / 3
    tau = np.zeros((n, DD[0] + 1, DD[1] + 1))
    tau[0:n:2, 0, 0] = pt1[0:n:2]
    tau[1:n:2, 0, 0] = pt2[1:n:2]
    tau[n - 1, 0, 0] = 1
    tau[tau > 1] = 1

    # Find the EC densities:
    u = df1 * t
    for d in range(1, np.max(DD) + 1):
        e_loop = np.min([np.min(DD), d]) + 1
        for e in range(0, e_loop):
            s1 = 0
            cons = -((d + e) / 2 + 1) * np.log(math.pi) + gammaln(d) + gammaln(e + 1)
            for k in colon(0, (d - 1 + e) / 2):
                j, i = np.meshgrid(np.arange(0, k + 1), np.arange(0, k + 1))
                if df2 == math.inf:
                    q1 = np.log(math.pi) / 2 - ((d + e - 1) / 2 + i + j) * np.log(2)
                else:
                    q1 = (
                        (df0 - 1 - d - e) * np.log(2)
                        + gammaln((df0 - d) / 2 + i)
                        + gammaln((df0 - e) / 2 + j)
                        - gammalni(df0 - d - e + i + j + k)
                        - ((d + e - 1) / 2 - k) * np.log(df2)
                    )
                q2 = (
                    cons
                    - gammalni(i + 1)
                    - gammalni(j + 1)
                    - gammalni(k - i - j + 1)
                    - gammalni(d - k - i + j)
                    - gammalni(e - k - j + i + 1)
                )
                s2 = np.sum(np.exp(q1 + q2))
                if s2 > 0:
                    s1 = s1 + (-1) ** k * u ** ((d + e - 1) / 2 - k) * s2

            if df2 == math.inf:
                s1 = s1 * np.exp(-u / 2)
            else:
                s1 = s1 * np.exp(-(df0 - 2) / 2 * np.log(1 + u / df2))

            if DD[0] >= DD[1]:
                tau[:, d, e] = s1
                if d <= np.min(DD):
                    tau[:, e, d] = s1
            else:
                tau[:, e, d] = s1
                if d <= np.min(DD):
                    tau[:, d, e] = s1

    # For multivariate statistics, add a sphere to the search region:
    a = np.zeros((2, np.max(nvar)))
    for k in range(0, 2):
        j = colon((nvar[k] - 1), 0, -2)
        a[k, j] = np.exp(
            j * np.log(2)
            + j / 2 * np.log(math.pi)
            + gammaln((nvar[k] + 1) / 2)
            - gammaln((nvar[k] + 1 - j) / 2)
            - gammaln(j + 1)
        )

    rho = np.zeros((n, Dlim[0] + 1, Dlim[1] + 1))

    for k in range(0, nvar[0]):
        for l in range(0, nvar[1]):
            rho = (
                rho
                + a[0, k] * a[1, l] * tau[:, k : Dlim[0] + k + 1, l : Dlim[1] + l + 1]
            )

    if is_tstat:
        if all(nvar == 1):
            t = np.r_[np.sqrt(t[0 : n - 1]), -np.sqrt(t)[::-1]]
            rho = np.r_[rho[0 : n - 1, :, :], rho[::-1, :, :]] / 2
            for i in range(0, D[0] + 1):
                for j in range(0, D[1] + 1):
                    rho[n - 1 + np.arange(0, n), i, j] = (
                        -((-1) ** (i + j)) * rho[n - 1 + np.arange(0, n), i, j]
                    )
            rho[n - 1 + np.arange(0, n), 0, 0] = rho[n - 1 + np.arange(0, n), 0, 0] + 1
            n = 2 * n - 1
        else:
            t = np.sqrt(t)

    # For scale space.
    if scale > 1:
        kappa = D[0] / 2
        tau = np.zeros(n, D[0] + 1)
        for d in range(0, D[0] + 1):
            s1 = 0
            for k in range(0, d / 2 + 1):
                s1 = (
                    s1 + (-1)
                    ^ k
                    / (1 - 2 * k)
                    * np.exp(
                        gammaln(d + 1)
                        - gammaln(k + 1)
                        - gammaln(d - 2 * k + 1)
                        + (1 / 2 - k) * np.log(kappa)
                        - k * np.log(4 * math.pi)
                    )
                    * rho[:, d + 1 - 2 * k, 1]
                )
            if d == 0:
                cons = np.log(scale)
            else:
                cons = (1 - 1 / scale ** d) / d
            tau[:, d] = rho[:, d, 1] * (1 + 1 / scale ** d) / 2 + s1 * cons
        rho[:, 0 : D[1], 0] = tau

    if D[1] == 0:
        d = D[0]
        if nconj > 1:
            # Conjunctions
            b = gamma((np.arange(1, d + 2) / 2)) / gamma(1 / 2)
            for i in range(0, d + 1):
                rho[:, i, 0] = rho[:, i, 0] / b[i]
            m1 = np.zeros((n, d + 1, d + 1))
            for i in range(0, d + 1):
                j = np.arange(i, d + 1)
                m1[:, i, j] = rho[:, j - i, 0]
            m2 = np.zeros(m1.shape)
            for _ in range(2, nconj + 1):
                for i in range(0, d + 1):
                    for j in range(0, d + 1):
                        m2[:, i, j] = np.sum(
                            rho[:, 0 : d - i + 1, 0] * m1[:, i : d + 1, j], axis=1
                        )
                m1 = m2
            for i in range(0, d + 1):
                rho[:, i, 0] = m1[:, 0, i] * b[i]
        if EC_file is not None:
            raise ValueError(
                "EC File support has not been implemented. We are not living in the 90s anymore."
            )

    if all(fwhm > 0):
        pval_rf = np.zeros(n)
        for i in range(0, D[0] + 1):
            for j in range(0, D[1] + 1):
                pval_rf = pval_rf + invol[0, i] * invol[1, j] * rho[:, i, j]
    else:
        pval_rf = math.inf

    # Bonferonni
    pt = rho[:, 0, 0]
    pval_bon = abs(np.prod(num_voxels)) * pt

    # Minimum of the two
    if type(pval_rf) != type(pval_bon):
        pval = np.minimum(pval_rf, pval_bon)
    else:
        if isinstance(pval_rf, float) and isinstance(pval_bon, float):
            pval = np.minimum(pval_rf, pval_bon)
        if isinstance(pval_rf, np.ndarray) and isinstance(pval_bon, np.ndarray):
            # this will ignore nan values in arrays while finding minimum
            pval_tmp = np.concatenate(
                (
                    pval_rf.reshape(pval_rf.shape[0], 1),
                    pval_bon.reshape(pval_bon.shape[0], 1),
                ),
                axis=1,
            )
            pval = np.nanmin(pval_tmp, axis=1)

    tlim = 1
    if p_val_peak.flatten()[0] <= tlim:
        peak_threshold = minterp1(pval, t, p_val_peak)
        if p_val_peak.size <= nprint:
            print(peak_threshold)
    else:
        # p_val_peak is treated as peak value
        P_val_peak = interp1(t, pval, p_val_peak)
        i = np.isnan(P_val_peak)
        P_val_peak[i] = is_tstat and (p_val_peak[i] < 0)
        peak_threshold = P_val_peak
        if p_val_peak.size <= nprint:
            print(P_val_peak)

    if np.all(fwhm <= 0) or np.any(num_voxels < 0):
        peak_threshold_1 = p_val_peak + float("nan")
        extent_threshold = p_val_extent + float("nan")
        extent_threshold_1 = extent_threshold + float("nan")
        return (
            peak_threshold,
            extent_threshold,
            peak_threshold_1,
            extent_threshold_1,
            t,
            rho,
        )

    # Cluster threshold:

    if cluster_threshold > tlim:
        tt = cluster_threshold
    else:
        # cluster_threshold is treated as a probability
        tt = minterp1(pt, t, cluster_threshold)

    d = np.sum(D)
    rhoD = interp1(t, rho[:, D[0], D[1]], tt)
    p = interp1(t, pt, tt)

    # Pre-selected peak
    pval = rho[:, D[0], D[1]] / rhoD
    if p_val_peak.flatten()[0] <= tlim:
        peak_threshold_1 = minterp1(pval, t, p_val_peak)
        if p_val_peak.size <= nprint:
            print(peak_threshold_1)
    else:
        # p_val_peak is treated as a peak value
        P_val_peak_1 = interp1(t, pval, p_val_peak)
        i = np.isnan(P_val_peak)
        P_val_peak_1[i] = is_tstat and (p_val_peak[i] < 0)
        peak_threshold_1 = P_val_peak_1
        if p_val_peak.size <= nprint:
            print(P_val_peak_1)

    if d == 0 or nconj > 1 or nvar[0] > 1 or scale > 1:
        extent_threshold = p_val_extent + float("nan")
        extent_threshold_1 = extent_threshold
        if p_val_extent.size <= nprint:
            print(extent_threshold)
            print(extent_threshold_1)
        return (
            peak_threshold,
            extent_threshold,
            peak_threshold_1,
            extent_threshold_1,
            t,
            rho,
        )

    # Expected number of clusters
    EL = invol[0, D[0]] * invol[1, D[1]] * rhoD
    cons = (
        gamma(d / 2 + 1)
        * (4 * np.log(2)) ** (d / 2)
        / fwhm[0] ** D[0]
        / fwhm[1] ** D[1]
        * rhoD
        / p
    )

    if df2 == math.inf and dfw1[0] == math.inf and dfw1[1] == math.inf:
        if p_val_extent.flatten()[0] <= tlim:
            pS = -np.log(1 - p_val_extent) / EL
            extent_threshold = (-np.log(pS)) ** (d / 2) / cons
            pS = -np.log(1 - p_val_extent)
            extent_threshold_1 = (-np.log(pS)) ** (d / 2) / cons
            if p_val_extent.size <= nprint:
                print(extent_threshold)
                print(extent_threshold_1)
        else:
            # p_val_extent is now treated as a spatial extent:
            pS = np.exp(-((p_val_extent * cons) ** (2 / d)))
            P_val_extent = 1 - np.exp(-pS * EL)
            extent_threshold = P_val_extent
            P_val_extent_1 = 1 - np.exp(-pS)
            extent_threshold_1 = P_val_extent_1
            if p_val_extent.size <= nprint:
                print(P_val_extent)
                print(P_val_extent_1)
    else:
        # Find dbn of S by taking logs then using fft for convolution:
        ny = 2 ** 12
        a = d / 2
        b2 = a * 10 * np.max([np.sqrt(2 / (np.min([df1 + df2, np.min(dfw1)]))), 1])
        if df2 < math.inf:
            b1 = a * np.log((1 - (1 - 0.000001) ** (2 / (df2 - d))) * df2 / 2)
        else:
            b1 = a * np.log(-np.log(1 - 0.000001))

        dy = (b2 - b1) / ny
        b1 = round(b1 / dy) * dy
        y = np.arange(0, ny) * dy + b1
        numrv = (
            1
            + (d + (D[0] > 0) + (D[1] > 0)) * (df2 < math.inf)
            + (D[0] * (dfw1[0] < math.inf) + (dfw2[0] < math.inf)) * (D[0] > 0)
            + (D[1] * (dfw1[1] < math.inf) + (dfw2[1] < math.inf)) * (D[1] > 0)
        )
        f = np.zeros((ny, numrv))
        if f.ndim == 1:
            f = np.expand_dims(f, axis=1)
        mu = np.zeros(numrv)
        if df2 < math.inf:
            # Density of log(Beta(1,(df2-d)/2)^(d/2)):
            yy = np.exp(y / a) / df2 * 2
            yy = yy * (yy < 1)
            f[:, 0] = (1 - yy) ** ((df2 - d) / 2 - 1) * ((df2 - d) / 2) * yy / a
            mu[0] = np.exp(
                gammaln(a + 1)
                + gammaln((df2 - d + 2) / 2)
                - gammaln((df2 + 2) / 2)
                + a * np.log(df2 / 2)
            )
        else:
            # Density of log(exp(1)^(d/2)):
            yy = np.exp(y / a)
            f[:, 0] = np.exp(-yy) * yy / a
            mu[0] = np.exp(gammaln(a + 1))

        nuv = np.array([])
        aav = np.array([])
        if df2 < math.inf:
            nuv = df2 + 2 - np.arange(1, d + 1)
            aav = np.ones((1, d)) * (-1 / 2)
            for k in range(0, 2):
                if D[k] > 0:
                    nuv = np.append(df1 + df2 - D[k], nuv)
                    aav = np.append(D[k] / 2, aav)

        for k in range(0, 2):
            if dfw1[k] < math.inf and D[k] > 0:
                if dfw1[k] > df_limit:
                    nuv = np.append(
                        nuv, dfw1[k] - dfw1[k] / dfw2[k] - np.arange(0, D[k])
                    )
                else:
                    nuv = np.append(
                        nuv, (dfw1[k] - dfw1[k] / dfw2[k]) * np.ones((1, D[k]))
                    )
                aav = np.append(aav, (1 / 2) * np.ones((1, D[k])))
            if dfw2[k] < math.inf:
                nuv = np.append(nuv, dfw2[k])
                aav = np.append(aav, -D[k] / 2)

        for i in range(0, numrv - 1):
            nu = nuv[i]
            aa = aav[i]
            yy = y / aa + np.log(nu)
            # Density of log((chi^2_nu/nu)^aa):
            f[:, i + 1] = np.exp(
                nu / 2 * yy - np.exp(yy) / 2 - (nu / 2) * np.log(2) - gammaln(nu / 2)
            ) / abs(aa)
            mu[i + 1] = np.exp(
                gammaln(nu / 2 + aa) - gammaln(nu / 2) - aa * np.log(nu / 2)
            )

        # Check: plot(y,f); sum(f*dy,1) should be 1

        omega = 2 * math.pi * np.arange(0, ny) / ny / dy
        shift = (np.cos(-b1 * omega) + np.sin(-b1 * omega) * 1j) * dy
        prodfft = np.prod(np.fft.fft(f, axis=0), axis=1) * shift ** (numrv - 1)

        # Density of Y=log(B^(d/2)*U^(d/2)/sqrt(det(Q)))
        ff = np.real(np.fft.ifft(prodfft, axis=0))
        # Check: plot(y,ff); sum(ff*dy) should be 1
        mu0 = np.prod(mu)
        # Check: plot(y,ff.*exp(y)); sum(ff.*exp(y)*dy.*(y<10)) should equal mu0
        alpha = (
            p
            / rhoD
            / mu0
            * fwhm[0] ** D[0]
            * fwhm[1] ** D[1]
            / (4 * np.log(2)) ** (d / 2)
        )

        # Integrate the density to get the p-value for one cluster:
        pS = np.cumsum(np.flip(ff)) * dy
        pS = np.flip(pS)

        # The number of clusters is Poisson with mean EL:
        pSmax = 1 - np.exp(-pS * EL)
        if p_val_extent.flatten()[0] <= tlim:
            yval = minterp1(-pSmax, y, -p_val_extent)
            # Spaytial extent is alpha*exp(Y) -dy/2 correction for mid-point rule:
            extent_threshold = alpha * np.exp(yval - dy / 2)
            # For a single cluster:
            yval = minterp1(-pS, y, -p_val_extent)
            extent_threshold_1 = alpha * np.exp(yval - dy / 2)
            if p_val_extent.size <= nprint:
                print(extent_threshold)
                print(extent_threshold_1)
        else:
            # p_val_extent is now treated as a spatial extent:
            logpval = np.log(p_val_extent / alpha + (p_val_extent <= 0)) + dy / 2
            P_val_extent = interp1(y, pSmax, logpval)
            extent_threshold = P_val_extent * (p_val_extent > 0) + (p_val_extent <= 0)
            # For a single cluster:
            P_val_extent_1 = interp1(y, pS, logpval)
            extent_threshold_1 = P_val_extent_1 * (p_val_extent > 0) + (
                p_val_extent <= 0
            )
            if p_val_extent.size <= nprint:
                print(P_val_extent)
                print(P_val_extent_1)

    return (
        peak_threshold,
        extent_threshold,
        peak_threshold_1,
        extent_threshold_1,
        t,
        rho,
    )


def peak_clus(
    self,
    thresh: float,
    reselspvert: Optional[np.ndarray] = None,
    edg: Optional[np.ndarray] = None,
) -> Tuple[dict, dict, np.ndarray]:
    """Finds peaks (local maxima) and clusters for surface data.
    Parameters
    ----------
    thresh : float,
        clusters are vertices where slm['t'][0,mask]>=thresh.
    reselspvert : numpy array of shape (v),
        resels per vertex, by default: np.ones(v).
    edg :  numpy array of shape (e,2), dtype=int,
        edge indices, by default computed from mesh_edges function.


    Returns
    -------
    peak : a dictionary with keys 't', 'vertid', 'clusid'.
        peak['t'] : numpy array of shape (np,1),
            array of peaks (local maxima).
        peak['vertid] : numpy array of shape (np,1),
            array of vertex id's (1-based).
        peak['clusid'] : numpy array of shape (np,1),
            array of cluster id's that contain the peak.
    clus : a dictionary with keys 'clusid', 'nverts', 'resels'.
        clus['clusid'] : numpy array of shape (nc,1),
            array of cluster id numbers.
        clus['nverts'] : numpy array of shape (nc,1),
            array of number of vertices in the cluster.
        clus['resels'] : numpy array of shape (nc,1),
            array of resels in the cluster.
    clusid : numpy array of shape (1,v),
        array of cluster id's for each vertex.
    """

    if self.k > 1 and self.df is None:
        raise ValueError("If k>1 then df must be defined.")

    if edg is None:
        edg = mesh_edges(self)

    l, v = np.shape(self.t)
    slm_t = copy.deepcopy(self.t)
    slm_t[0, ~self.mask.astype(bool)] = slm_t[0, :].min()
    t1 = slm_t[0, edg[:, 0]]
    t2 = slm_t[0, edg[:, 1]]
    islm = np.ones((1, v))
    islm[0, edg[t1 < t2, 0]] = 0
    islm[0, edg[t2 < t1, 1]] = 0
    lmvox = np.argwhere(islm)[:, 1] + 1
    excurset = np.array(slm_t[0, :] >= thresh, dtype=int)
    n = excurset.sum()

    if n < 1:
        peak = []
        clus = []
        clusid = []
        return peak, clus, clusid

    voxid = np.cumsum(excurset)
    edg = voxid[edg[np.all(excurset[edg], 1), :]]
    nf = np.arange(1, n + 1)

    # Find cluster id's in nf (from Numerical Recipes in C, page 346):
    for el in range(1, edg.shape[0] + 1):
        j = edg[el - 1, 0]
        k = edg[el - 1, 1]
        while nf[j - 1] != j:
            j = nf[j - 1]
        while nf[k - 1] != k:
            k = nf[k - 1]
        if j != k:
            nf[j - 1] = k

    for j in range(1, n + 1):
        while nf[j - 1] != nf[nf[j - 1] - 1]:
            nf[j - 1] = nf[nf[j - 1] - 1]

    vox = np.argwhere(excurset) + 1
    ivox = np.argwhere(np.in1d(vox, lmvox)) + 1
    clmid = nf[ivox - 1]
    uclmid, iclmid, jclmid = np.unique(clmid, return_index=True, return_inverse=True)
    iclmid = iclmid + 1
    jclmid = jclmid + 1
    ucid = np.unique(nf)
    nclus = len(ucid)
    # implementing matlab's histc function ###
    bin_edges = np.r_[-np.Inf, 0.5 * (ucid[:-1] + ucid[1:]), np.Inf]
    ucvol, ucvol_edges = np.histogram(nf, bin_edges)

    if reselspvert is None:
        reselsvox = np.ones(np.shape(vox))
    else:
        reselsvox = reselspvert[vox - 1]

    # calling matlab-python version for scipy's interp1d
    nf1 = interp1(np.append(0, ucid), np.arange(0, nclus + 1), nf, kind="nearest")

    # if k>1, find volume of cluster in added sphere
    if self.k is None or self.k == 1:
        ucrsl = np.bincount(nf1.astype(int), reselsvox.flatten())
    if self.k == 2:
        if l == 1:
            ndf = np.array([self.df]).size
            r = 2 * np.arccos((thresh / slm_t[0, vox - 1]) ** (float(1) / ndf))
        else:
            r = 2 * np.arccos(
                np.sqrt(
                    (thresh - slm_t[1, vox - 1])
                    * (thresh >= slm_t[1, vox - 1])
                    / (slm_t[0, vox - 1] - slm_t[1, vox - 1])
                )
            )
        ucrsl = np.bincount(nf1.astype(int), (r.T * reselsvox.T).flatten())
    if self.k == 3:
        if l == 1:
            ndf = np.array([self.df]).size
            r = 2 * math.pi * (1 - (thresh / slm_t[0, vox - 1]) ** (float(1) / ndf))
        else:
            nt = 20
            theta = (np.arange(1, nt + 1, 1) - 1 / 2) / nt * math.pi / 2
            s = (np.cos(theta) ** 2 * slm_t[1, vox - 1]).T
            if l == 3:
                s = s + ((np.sin(theta) ** 2) * slm_t[2, vox - 1]).T
            r = (
                2
                * math.pi
                * (
                    1
                    - np.sqrt(
                        (thresh - s)
                        * (thresh >= s)
                        / (np.ones((nt, 1)) * slm_t[0, vox - 1].T - s)
                    )
                ).mean(axis=0)
            )
        ucrsl = np.bincount(nf1.astype(int), (r.T * reselsvox.T).flatten())

    # and their ranks (in ascending order)
    iucrls = sorted(range(len(ucrsl[1:])), key=lambda k: ucrsl[1:][k])
    rankrsl = np.zeros((1, nclus))
    rankrsl[0, iucrls] = np.arange(nclus, 0, -1)

    lmid = lmvox[ismember(lmvox, vox)[0]]

    varA = slm_t[0, (lmid - 1)]
    varB = lmid
    varC = rankrsl[0, jclmid - 1]
    varALL = np.concatenate(
        (
            varA.reshape(len(varA), 1),
            varB.reshape(len(varB), 1),
            varC.reshape(len(varC), 1),
        ),
        axis=1,
    )
    lm = np.flipud(
        varALL[
            varALL[:, 0].argsort(),
        ]
    )
    varNEW = np.concatenate(
        (rankrsl.T, ucvol.reshape(len(ucvol), 1), ucrsl.reshape(len(ucrsl), 1)[1:]),
        axis=1,
    )
    cl = varNEW[
        varNEW[:, 0].argsort(),
    ]
    clusid = np.zeros((1, v))
    clusid[0, (vox - 1).T] = interp1(
        np.append(0, ucid), np.append(0, rankrsl), nf, kind="nearest"
    )
    peak = {}
    peak["t"] = lm[:, 0].reshape(len(lm[:, 0]), 1)
    peak["vertid"] = lm[:, 1].reshape(len(lm[:, 1]), 1)
    peak["clusid"] = lm[:, 2].reshape(len(lm[:, 2]), 1)
    clus = {}
    clus["clusid"] = cl[:, 0].reshape(len(cl[:, 0]), 1)
    clus["nverts"] = cl[:, 1].reshape(len(cl[:, 1]), 1)
    clus["resels"] = cl[:, 2].reshape(len(cl[:, 2]), 1)

    return peak, clus, clusid


def compute_resels(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """SurfStatResels of surface or volume data inside a mask.

    Parameters
    ----------
    slm : a dictionary with keys 'lat' or 'tri' and, optionally, 'resl'.
        slm['lat'] : a 3D numpy array of 1's and 0's.
        slm['tri'] : a 2D numpy array of shape (t, 3).
            Contains triangles of a surface. slm['tri'].max() is equal to the
            number of vertices.
        slm['resl'] : a 2D numpy array of shape (e, k).
            Sum over observations of squares of differences of normalized
            residuals along each edge.
    mask : a 1D numpy array of shape (v), dtype 'bool'.
        v must be equal to int(slm['tri'].max()).
        Contains 1's and 0's (1's are included and 0's are excluded).

    Returns
    -------
    resels : a 2D numpy array of shape (1, (D+1)).
        Array of 0,...,D dimensional resels of the mask, EC of the mask
        if slm['resl'] is not given.
    reselspvert : a 1D numpy array of shape (v).
        Array of D-dimensional resels per mask vertex.
    edg : a 2D numpy array of shape (e, 2).
        Array of edge indices.
    """

    def pacos(x: ArrayLike) -> np.ndarray:
        return np.arccos(np.minimum(np.abs(x), 1) * np.sign(x))

    if self.tri is not None:
        # Get unique edges. Subtract 1 from edges to conform to Python's
        # counting from 0 - RV
        tri = np.sort(self.tri) - 1
        edg = np.unique(
            np.vstack((tri[:, (0, 1)], tri[:, (0, 2)], tri[:, (1, 2)])), axis=0
        )

        # If no mask is provided, create one with all included vertices set to
        # 1. If mask is provided, simply grab the number of vertices from mask.
        if self.mask is None:
            v = np.amax(edg) + 1
            self.mask = np.full(v, False)
            self.mask[edg] = True
        else:
            # if np.ndim(mask) > 1:
            #    mask = np.squeeze(mask)
            #    if mask.shape[0] > 1:
            #        mask = mask.T
            v = self.mask.size

        # Compute the Lipschitz–Killing curvatures (LKC)
        m = np.sum(self.mask)
        if self.resl is not None:
            lkc = np.zeros((3, 3))
        else:
            lkc = np.zeros((1, 3))
        lkc[0, 0] = m

        # LKC of edges
        maskedg = np.all(self.mask[edg], axis=1)
        lkc[0, 1] = np.sum(maskedg)

        if self.resl is not None:
            r1 = np.mean(np.sqrt(self.resl[maskedg, :]), axis=1)
            lkc[1, 1] = np.sum(r1)
        # LKC of triangles
        # Made an adjustment from the MATLAB implementation:
        # The reselspvert computation is included in the if-statement.
        # MATLAB errors when the if statement is false as variable r2 is not
        # defined during the computation of reselspvert. - RV
        masktri = np.all(self.mask[tri], 1)
        lkc[0, 2] = np.sum(masktri)
        if self.resl is not None:
            loc = row_ismember(tri[masktri, :][:, [0, 1]], edg)
            l12 = self.resl[loc, :]
            loc = row_ismember(tri[masktri, :][:, [0, 2]], edg)
            l13 = self.resl[loc, :]
            loc = row_ismember(tri[masktri, :][:, [1, 2]], edg)
            l23 = self.resl[loc, :]
            a = np.fmax(4 * l12 * l13 - (l12 + l13 - l23) ** 2, 0)
            r2 = np.mean(np.sqrt(a), axis=1) / 4
            lkc[1, 2] = (
                np.sum(np.mean(np.sqrt(l12) + np.sqrt(l13) + np.sqrt(l23), axis=1)) / 2
            )
            lkc[2, 2] = np.nansum(r2, axis=0)

            # Compute resels per mask vertex
            reselspvert = np.zeros(v)
            for j in range(0, 3):
                reselspvert = reselspvert + np.bincount(
                    tri[masktri, j], weights=r2, minlength=v
                )
            D = 2
            reselspvert = reselspvert / (D + 1) / np.sqrt(4 * np.log(2)) ** D
        else:
            reselspvert = None

    if self.lat is not None:
        edg = mesh_edges(self)
        # The lattice is filled with 5 alternating tetrahedra per cube
        I, J, K = np.shape(self.lat)
        IJ = I * J
        i, j = np.meshgrid(range(1, I + 1), range(1, J + 1))
        i = np.squeeze(np.reshape(i, (-1, 1)))
        j = np.squeeze(np.reshape(j, (-1, 1)))

        c1 = np.argwhere((((i + j) % 2) == 0) & (i < I) & (j < J))
        c2 = np.argwhere((((i + j) % 2) == 0) & (i > 1) & (j < J))
        c11 = np.argwhere((((i + j) % 2) == 0) & (i == I) & (j < J))
        c21 = np.argwhere((((i + j) % 2) == 0) & (i == I) & (j > 1))
        c12 = np.argwhere((((i + j) % 2) == 0) & (i < I) & (j == J))
        c22 = np.argwhere((((i + j) % 2) == 0) & (i > 1) & (j == J))

        # outcome is 1 lower than MATLAB due to 0-1 counting difference. - RV
        d1 = np.argwhere((((i + j) % 2) == 1) & (i < I) & (j < J)) + IJ
        d2 = np.argwhere((((i + j) % 2) == 1) & (i > 1) & (j < J)) + IJ

        tri1 = cat(
            (
                cat((c1, c1 + 1, c1 + 1 + I), axis=1),
                cat((c1, c1 + I, c1 + 1 + I), axis=1),
                cat((c2 - 1, c2, c2 - 1 + I), axis=1),
                cat((c2, c2 - 1 + I, c2 + I), axis=1),
            ),
            axis=0,
        )
        tri2 = cat(
            (
                cat((c1, c1 + 1, c1 + 1 + IJ), axis=1),
                cat((c1, c1 + IJ, c1 + 1 + IJ), axis=1),
                cat((c1, c1 + I, c1 + I + IJ), axis=1),
                cat((c1, c1 + IJ, c1 + I + IJ), axis=1),
                cat((c1, c1 + 1 + I, c1 + 1 + IJ), axis=1),
                cat((c1, c1 + 1 + I, c1 + I + IJ), axis=1),
                cat((c1, c1 + 1 + IJ, c1 + I + IJ), axis=1),
                cat((c1 + 1 + I, c1 + 1 + IJ, c1 + I + IJ), axis=1),
                cat((c2 - 1, c2, c2 - 1 + IJ), axis=1),
                cat((c2, c2 - 1 + IJ, c2 + IJ), axis=1),
                cat((c2 - 1, c2 - 1 + I, c2 - 1 + IJ), axis=1),
                cat((c2 - 1 + I, c2 - 1 + IJ, c2 - 1 + I + IJ), axis=1),
                cat((c2, c2 - 1 + I, c2 + I + IJ), axis=1),
                cat((c2, c2 - 1 + IJ, c2 + I + IJ), axis=1),
                cat((c2, c2 - 1 + I, c2 - 1 + IJ), axis=1),
                cat((c2 - 1 + I, c2 - 1 + IJ, c2 + I + IJ), axis=1),
                cat((c11, c11 + I, c11 + I + IJ), axis=1),
                cat((c11, c11 + IJ, c11 + I + IJ), axis=1),
                cat((c21 - I, c21, c21 - I + IJ), axis=1),
                cat((c21, c21 - I + IJ, c21 + IJ), axis=1),
                cat((c12, c12 + 1, c12 + 1 + IJ), axis=1),
                cat((c12, c12 + IJ, c12 + 1 + IJ), axis=1),
                cat((c22 - 1, c22, c22 - 1 + IJ), axis=1),
                cat((c22, c22 - 1 + IJ, c22 + IJ), axis=1),
            ),
            axis=0,
        )
        tri3 = cat(
            (
                cat((d1, d1 + 1, d1 + 1 + I), axis=1),
                cat((d1, d1 + I, d1 + 1 + I), axis=1),
                cat((d2 - 1, d2, d2 - 1 + I), axis=1),
                cat((d2, d2 - 1 + I, d2 + I), axis=1),
            ),
            axis=0,
        )
        tet1 = cat(
            (
                cat((c1, c1 + 1, c1 + 1 + I, c1 + 1 + IJ), axis=1),
                cat((c1, c1 + I, c1 + 1 + I, c1 + I + IJ), axis=1),
                cat((c1, c1 + 1 + I, c1 + 1 + IJ, c1 + I + IJ), axis=1),
                cat((c1, c1 + IJ, c1 + 1 + IJ, c1 + I + IJ), axis=1),
                cat((c1 + 1 + I, c1 + 1 + IJ, c1 + I + IJ, c1 + 1 + I + IJ), axis=1),
                cat((c2 - 1, c2, c2 - 1 + I, c2 - 1 + IJ), axis=1),
                cat((c2, c2 - 1 + I, c2 + I, c2 + I + IJ), axis=1),
                cat((c2, c2 - 1 + I, c2 - 1 + IJ, c2 + I + IJ), axis=1),
                cat((c2, c2 - 1 + IJ, c2 + IJ, c2 + I + IJ), axis=1),
                cat((c2 - 1 + I, c2 - 1 + IJ, c2 - 1 + I + IJ, c2 + I + IJ), axis=1),
            ),
            axis=0,
        )

        v = np.int(np.round(np.sum(self.lat)))
        if self.mask is None:
            self.mask = np.ones(v, dtype=bool)

        reselspvert = np.zeros(v)
        vs = np.cumsum(np.squeeze(np.sum(np.sum(self.lat, axis=0), axis=0)))
        vs = cat((np.zeros(1), vs, np.expand_dims(vs[K - 1], axis=0)), axis=0)
        vs = vs.astype(int)
        es = 0
        lat = np.zeros((I, J, 2))
        lat[:, :, 0] = self.lat[:, :, 0]
        lkc = np.zeros((4, 4))
        for k in range(0, K):
            f = (k + 1) % 2
            if k < (K - 1):
                lat[:, :, f] = self.lat[:, :, k + 1]
            else:
                lat[:, :, f] = np.zeros((I, J))
            vid = (np.cumsum(lat.flatten("F")) * np.reshape(lat.T, -1)).astype(int)
            if f:
                edg1 = (
                    edg[
                        np.logical_and(
                            edg[:, 0] > (vs[k] - 1), edg[:, 0] <= (vs[k + 1] - 1)
                        ),
                        :,
                    ]
                    - vs[k]
                )
                edg2 = (
                    edg[
                        np.logical_and(
                            edg[:, 0] > (vs[k] - 1), edg[:, 1] <= (vs[k + 2] - 1)
                        ),
                        :,
                    ]
                    - vs[k]
                )
                # Added a -1 - RV
                tri = (
                    cat(
                        (
                            vid[
                                tri1[
                                    np.all(
                                        np.reshape(lat.flatten("F")[tri1], tri1.shape),
                                        1,
                                    ),
                                    :,
                                ]
                            ],
                            vid[
                                tri2[
                                    np.all(
                                        np.reshape(lat.flatten("F")[tri2], tri2.shape),
                                        1,
                                    ),
                                    :,
                                ]
                            ],
                        ),
                        axis=0,
                    )
                    - 1
                )
                mask1 = self.mask[np.arange(vs[k], vs[k + 2])]
            else:
                edg1 = cat(
                    (
                        edg[
                            np.logical_and(
                                edg[:, 0] > (vs[k] - 1), edg[:, 1] <= (vs[k + 1] - 1)
                            ),
                            :,
                        ]
                        - vs[k]
                        + vs[k + 2]
                        - vs[k + 1],
                        cat(
                            (
                                np.expand_dims(
                                    edg[
                                        np.logical_and(
                                            edg[:, 0] <= (vs[k + 1] - 1),
                                            edg[:, 1] > (vs[k + 1] - 1),
                                        ),
                                        1,
                                    ]
                                    - vs[k + 1],
                                    axis=1,
                                ),
                                np.expand_dims(
                                    edg[
                                        np.logical_and(
                                            edg[:, 0] <= (vs[k + 1] - 1),
                                            edg[:, 1] > (vs[k + 1] - 1),
                                        ),
                                        0,
                                    ]
                                    - vs[k]
                                    + vs[k + 2]
                                    - vs[k + 1],
                                    axis=1,
                                ),
                            ),
                            axis=1,
                        ),
                    ),
                    axis=0,
                )
                edg2 = cat(
                    (
                        edg1,
                        edg[
                            np.logical_and(
                                edg[:, 0] > (vs[k + 1] - 1),
                                edg[:, 1] <= (vs[k + 2] - 1),
                            ),
                            :,
                        ]
                        - vs[k + 1],
                    ),
                    axis=0,
                )
                # Added a -1 - RV
                tri = (
                    cat(
                        (
                            vid[tri3[np.all(lat.flatten("F")[tri3], axis=1), :]],
                            vid[tri2[np.all(lat.flatten("F")[tri2], axis=1), :]],
                        ),
                        axis=0,
                    )
                    - 1
                )
                mask1 = cat(
                    (
                        self.mask[np.arange(vs[k + 1], vs[k + 2])],
                        self.mask[np.arange(vs[k], vs[k + 1])],
                    )
                )
            # Added a -1 -RV
            tet = vid[tet1[np.all(lat.flatten("F")[tet1], axis=1), :]] - 1
            m1 = np.max(edg2[:, 0])
            ue = edg2[:, 0] + m1 * (edg2[:, 1] - 1)
            e = edg2.shape[0]
            ae = np.arange(0, e)
            if e < 2 ** 31:
                sparsedg = csr_matrix(
                    (ae, (ue, np.zeros(ue.shape, dtype=int))), dtype=np.int
                )
                sparsedg.eliminate_zeros()
            ##
            lkc1 = np.zeros((4, 4))
            lkc1[0, 0] = np.sum(self.mask[np.arange(vs[k], vs[k + 1])])

            # LKC of edges
            maskedg = np.all(mask1[edg1], axis=1)

            lkc1[0, 1] = np.sum(maskedg)
            if self.resl is not None:
                r1 = np.mean(np.sqrt(self.resl[np.argwhere(maskedg) + es, :]), axis=1)
                lkc1[1, 1] = np.sum(r1)

            # LKC of triangles
            masktri = np.all(mask1[tri], axis=1).flatten()
            lkc1[0, 2] = np.sum(masktri)
            if self.resl is not None:
                if all(masktri == False):
                    # Set these variables to empty arrays to match the MATLAB
                    # implementation.
                    lkc1[1, 2] = 0
                    lkc1[2, 2] = 0
                else:
                    if e < 2 ** 31:
                        l12 = self.resl[
                            sparsedg[
                                tri[masktri, 0] + m1 * (tri[masktri, 1] - 1), 0
                            ].toarray()
                            + es,
                            :,
                        ]
                        l13 = self.resl[
                            sparsedg[
                                tri[masktri, 0] + m1 * (tri[masktri, 2] - 1), 0
                            ].toarray()
                            + es,
                            :,
                        ]
                        l23 = self.resl[
                            sparsedg[
                                tri[masktri, 1] + m1 * (tri[masktri, 2] - 1), 0
                            ].toarray()
                            + es,
                            :,
                        ]
                    else:
                        l12 = self.resl[
                            interp1(
                                ue,
                                ae,
                                tri[masktri, 0] + m1 * (tri[masktri, 1] - 1),
                                kind="nearest",
                            )
                            + es,
                            :,
                        ]
                        l13 = self.resl[
                            interp1(
                                ue,
                                ae,
                                tri[masktri, 0] + m1 * (tri[masktri, 2] - 1),
                                kind="nearest",
                            )
                            + es,
                            :,
                        ]
                        l23 = self.resl[
                            interp1(
                                ue,
                                ae,
                                tri[masktri, 1] + m1 * (tri[masktri, 2] - 1),
                                kind="nearest",
                            )
                            + es,
                            :,
                        ]
                    a = np.fmax(4 * l12 * l13 - (l12 + l13 - l23) ** 2, 0)
                    r2 = np.mean(np.sqrt(a), axis=1) / 4
                    lkc1[1, 2] = (
                        np.sum(
                            np.mean(np.sqrt(l12) + np.sqrt(l13) + np.sqrt(l23), axis=1)
                        )
                        / 2
                    )
                    lkc1[2, 2] = np.sum(r2)

                # The following if-statement has nargout >=2 in MATLAB,
                # but there's no Python equivalent so ignore that. - RV
                if K == 1:
                    for j in range(0, 3):
                        if f:
                            v1 = tri[masktri, j] + vs[k]
                        else:
                            v1 = tri[masktri, j] + vs[k + 1]
                            v1 = v1 - int(vs > vs[k + 2]) * (vs[k + 2] - vs[k])
                        reselspvert += np.bincount(v1, r2, v)

            # LKC of tetrahedra
            masktet = np.all(mask1[tet], axis=1).flatten()
            lkc1[0, 3] = np.sum(masktet)
            if self.resl is not None and k < (K - 1):
                if e < 2 ** 31:
                    l12 = self.resl[
                        (
                            sparsedg[
                                tet[masktet, 0] + m1 * (tet[masktet, 1] - 1), 0
                            ].toarray()
                            + es
                        ).tolist(),
                        :,
                    ]
                    l13 = self.resl[
                        (
                            sparsedg[
                                tet[masktet, 0] + m1 * (tet[masktet, 2] - 1), 0
                            ].toarray()
                            + es
                        ).tolist(),
                        :,
                    ]
                    l23 = self.resl[
                        (
                            sparsedg[
                                tet[masktet, 1] + m1 * (tet[masktet, 2] - 1), 0
                            ].toarray()
                            + es
                        ).tolist(),
                        :,
                    ]
                    l14 = self.resl[
                        (
                            sparsedg[
                                tet[masktet, 0] + m1 * (tet[masktet, 3] - 1), 0
                            ].toarray()
                            + es
                        ).tolist(),
                        :,
                    ]
                    l24 = self.resl[
                        (
                            sparsedg[
                                tet[masktet, 1] + m1 * (tet[masktet, 3] - 1), 0
                            ].toarray()
                            + es
                        ).tolist(),
                        :,
                    ]
                    l34 = self.resl[
                        (
                            sparsedg[
                                tet[masktet, 2] + m1 * (tet[masktet, 3] - 1), 0
                            ].toarray()
                            + es
                        ).tolist(),
                        :,
                    ]
                else:
                    l12 = self.resl[
                        interp1(
                            ue,
                            ae,
                            tet[masktet, 0] + m1 * (tet[masktet, 1] - 1),
                            kind="nearest",
                        )
                        + es,
                        :,
                    ]
                    l13 = self.resl[
                        interp1(
                            ue,
                            ae,
                            tet[masktet, 0] + m1 * (tet[masktet, 2] - 1),
                            kind="nearest",
                        )
                        + es,
                        :,
                    ]
                    l23 = self.resl[
                        interp1(
                            ue,
                            ae,
                            tet[masktet, 1] + m1 * (tet[masktet, 2] - 1),
                            kind="nearest",
                        )
                        + es,
                        :,
                    ]
                    l14 = self.resl[
                        interp1(
                            ue,
                            ae,
                            tet[masktet, 0] + m1 * (tet[masktet, 3] - 1),
                            kind="nearest",
                        )
                        + es,
                        :,
                    ]
                    l24 = self.resl[
                        interp1(
                            ue,
                            ae,
                            tet[masktet, 1] + m1 * (tet[masktet, 3] - 1),
                            kind="nearest",
                        )
                        + es,
                        :,
                    ]
                    l34 = self.resl[
                        interp1(
                            ue,
                            ae,
                            tet[masktet, 2] + m1 * (tet[masktet, 3] - 1),
                            kind="nearest",
                        )
                        + es,
                        :,
                    ]
                a4 = np.fmax(4 * l12 * l13 - (l12 + l13 - l23) ** 2, 0)
                a3 = np.fmax(4 * l12 * l14 - (l12 + l14 - l24) ** 2, 0)
                a2 = np.fmax(4 * l13 * l14 - (l13 + l14 - l34) ** 2, 0)
                a1 = np.fmax(4 * l23 * l24 - (l23 + l24 - l34) ** 2, 0)

                d12 = 4 * l12 * l34 - (l13 + l24 - l23 - l14) ** 2
                d13 = 4 * l13 * l24 - (l12 + l34 - l23 - l14) ** 2
                d14 = 4 * l14 * l23 - (l12 + l34 - l24 - l13) ** 2

                h = np.logical_or(a1 <= 0, a2 <= 0)
                delta12 = np.sum(
                    np.mean(
                        np.sqrt(l34)
                        * pacos(
                            (d12 - a1 - a2) / np.sqrt(a1 * a2 + h) / 2 * (1 - h) + h
                        ),
                        axis=1,
                    )
                )
                h = np.logical_or(a1 <= 0, a3 <= 0)
                delta13 = np.sum(
                    np.mean(
                        np.sqrt(l24)
                        * pacos(
                            (d13 - a1 - a3) / np.sqrt(a1 * a3 + h) / 2 * (1 - h) + h
                        ),
                        axis=1,
                    )
                )
                h = np.logical_or(a1 <= 0, a4 <= 0)
                delta14 = np.sum(
                    np.mean(
                        np.sqrt(l23)
                        * pacos(
                            (d14 - a1 - a4) / np.sqrt(a1 * a4 + h) / 2 * (1 - h) + h
                        ),
                        axis=1,
                    )
                )
                h = np.logical_or(a2 <= 0, a3 <= 0)
                delta23 = np.sum(
                    np.mean(
                        np.sqrt(l14)
                        * pacos(
                            (d14 - a2 - a3) / np.sqrt(a2 * a3 + h) / 2 * (1 - h) + h
                        ),
                        axis=1,
                    )
                )
                h = np.logical_or(a2 <= 0, a4 <= 0)
                delta24 = np.sum(
                    np.mean(
                        np.sqrt(l13)
                        * pacos(
                            (d13 - a2 - a4) / np.sqrt(a2 * a4 + h) / 2 * (1 - h) + h
                        ),
                        axis=1,
                    )
                )
                h = np.logical_or(a3 <= 0, a4 <= 0)
                delta34 = np.sum(
                    np.mean(
                        np.sqrt(l12)
                        * pacos(
                            (d12 - a3 - a4) / np.sqrt(a3 * a4 + h) / 2 * (1 - h) + h
                        ),
                        axis=1,
                    )
                )

                r3 = np.squeeze(
                    np.mean(
                        np.sqrt(
                            np.fmax(
                                (4 * a1 * a2 - (a1 + a2 - d12) ** 2)
                                / (l34 + (l34 <= 0))
                                * (l34 > 0),
                                0,
                            )
                        ),
                        axis=1,
                    )
                    / 48
                )

                lkc1[1, 3] = (
                    delta12 + delta13 + delta14 + delta23 + delta24 + delta34
                ) / (2 * np.pi)
                lkc1[2, 3] = (
                    np.sum(
                        np.mean(
                            np.sqrt(a1) + np.sqrt(a2) + np.sqrt(a3) + np.sqrt(a4),
                            axis=1,
                        )
                    )
                    / 8
                )
                lkc1[3, 3] = np.sum(r3)

                # Original MATLAB code has a if nargout>=2 here, ignore it
                # as no equivalent exists in Python - RV.
                for j in range(0, 4):
                    if f:
                        v1 = tet[masktet, j] + vs[k]
                    else:
                        v1 = tet[masktet, j] + vs[k + 1]
                        v1 = v1 - (v1 > (vs[k + 2] - 1)) * (vs[k + 2] - vs[k])
                    if np.ndim(r3) == 0:
                        r3 = r3.tolist()
                        r3 = [r3]
                    reselspvert += np.bincount(v1, r3, v)
            lkc = lkc + lkc1
            es = es + edg1.shape[0]

        # Original MATLAB code has a if nargout>=2 here,
        # ignore it as no equivalent exists in Python - RV.
        D = 2 + (K > 1)
        reselspvert = reselspvert / (D + 1) / np.sqrt(4 * np.log(2)) ** D

    # Compute resels - RV
    D1 = lkc.shape[0] - 1
    D2 = lkc.shape[1] - 1
    tpltz = toeplitz((-1) ** (np.arange(0, D1 + 1)), (-1) ** (np.arange(0, D2 + 1)))
    lkcs = np.sum(tpltz * lkc, axis=1).T
    lkcs = np.trim_zeros(lkcs, trim="b")
    lkcs = np.atleast_2d(lkcs)
    D = lkcs.shape[1] - 1
    resels = lkcs / np.sqrt(4 * np.log(2)) ** np.arange(0, D + 1)

    return resels, reselspvert, edg



"""Data generation for stat_threshold unit tests."""

import pickle

import numpy as np
from sklearn.model_selection import ParameterGrid

from brainstat.stats._multiple_comparisons import stat_threshold
from brainstat.tests.testutil import datadir


def generate_stat_threshold_out(I):
    D = {}
    (
        D["peak_threshold"],
        D["extent_threshold"],
        D["peak_threshold_1"],
        D["extent_threshold_1"],
        D["t"],
        D["rho"],
    ) = stat_threshold(
        I["search_volume"],
        I["num_voxels"],
        I["fwhm"],
        I["df"],
        I["p_val_peak"],
        I["cluster_threshold"],
        I["p_val_extent"],
        I["nconj"],
        I["nvar"],
        None,
        None,
        I["nprint"],
    )
    return D


def params2files(I, D, test_num):
    """Converts params to input/output files"""
    basename = "xthresh"
    fin_name = datadir(basename + "_" + f"{test_num:02d}" + "_IN.pkl")
    fout_name = datadir(basename + "_" + f"{test_num:02d}" + "_OUT.pkl")
    with open(fin_name, "wb") as f:
        pickle.dump(I, f, protocol=4)
    with open(fout_name, "wb") as g:
        pickle.dump(D, g, protocol=4)
    return


def generate_test_data():
    np.random.seed(0)

    mygrid = [
        {
            "search_volume": [
                np.random.rand() * 5,
                np.random.randint(1, 20, size=(3,)).tolist(),
                np.random.rand(2, 2) * 20,
            ],
            "num_voxels": [
                int(1),
                np.random.randint(1, 10, size=(3,)).tolist(),
            ],
            "fwhm": [0.0, np.random.rand() * 5],
            "df": [5, np.random.randint(10, size=(2, 2))],
            "p_val_peak": [
                0.05,
                np.random.rand(
                    4,
                ).tolist(),
            ],
            "cluster_threshold": [0.001],
            "p_val_extent": [0.05],
            "nconj": [0.5],
            "nvar": [1],
            "nprint": [0],
        }
    ]

    myparamgrid = ParameterGrid(mygrid)

    # Here wo go!
    # Tests 1-48 : search_volume -> float, list, 2D array
    # num_voxel -> int, list of ints,
    # fwhm -> 0, float, df -> int, 2D array of type ints,
    # p_val_peak --> 0.05, list of floats
    # cluster_threshold, p_val_extent, nconj, nvar, nprint -> default values

    test_num = 0
    for params in myparamgrid:
        test_num += 1
        I = {}
        for key in params.keys():
            I[key] = params[key]
        D = generate_stat_threshold_out(I)
        params2files(I, D, test_num)


if __name__ == "__main__":
    generate_test_data()



import warnings
from copy import deepcopy
from typing import Optional, Tuple, Union

import numpy as np
import numpy.linalg as la
from brainspace.mesh.mesh_elements import get_cells
from brainspace.vtk_interface.wrappers.data_object import BSPolyData

from brainstat.mesh.utils import mesh_edges
from brainstat.stats.terms import FixedEffect, MixedEffect


def _linear_model(self, Y: Union[np.ndarray, FixedEffect]) -> None:
    """Fits linear mixed effects models to surface data and estimates resels.

    Parameters
    ----------
    self : brainstat.stats.SLM.SLM
        Initialized SLM object.
    Y : numpy array, brainstat.stats.terms.FixedEffect
        Input data of shape (samples, vertices, features).

    """

    # PATCHWORK FIX: Y is modified in place a lot which also modifies it outside
    # the scope of this function. Lets just make a deep copy.
    # TODO: Stop modifying in place.
    Y_copy = deepcopy(Y)

    if isinstance(Y_copy, FixedEffect):
        Y_copy = Y_copy.m.to_numpy()
    Y_copy = np.atleast_3d(Y_copy)

    n_samples = Y.shape[0]

    self.X, self.V = _get_design_matrix(self, n_samples)
    _check_constant_term(self.X)

    self.df = n_samples - la.matrix_rank(self.X)
    residuals, self.V, self.coef, self.SSE, self.r, self.dr = _run_linear_model(
        self, Y_copy
    )

    if self.surf is not None:
        self.resl, mesh_connections = _compute_resls(self, residuals)
        key = list(mesh_connections.keys())[0]
        setattr(self, key, mesh_connections[key])


def _run_linear_model(
    self, Y: np.ndarray
) -> Tuple[
    np.ndarray,
    Optional[np.ndarray],
    np.ndarray,
    np.ndarray,
    Optional[np.ndarray],
    Optional[np.ndarray],
]:
    """Runs a linear model and returns relevant parameters.

    Parameters
    ----------
    Y : numpy.array
        Response variable matrix

    Returns
    -------
    numpy.array
        Model residuals.
    numpy.array
        Variance matrix bases.
    numpy.array
        Model coefficients.
    numpy.array
        Sum of squared errors.
    numpy.array, None
        Matrix of coefficients of the components of slm.V divided by their sum.
        None is returned for fixed effects models.
    numpy.array, None
        Vector of increments in slm.r None is returned for fixed effects models.

    Raises
    ------
    ValueError
        Returns an error if a multivariate mixed effects model is requested.
    """
    n_random_effects = _get_n_random_effects(self)
    r = None
    dr = None

    if Y.shape[2] == 1:  # Univariate
        Y = Y[:, :, 0]

        if n_random_effects == 1:  # Fixed effects
            residuals, V, coef, SSE = _model_univariate_fixed_effects(self, Y)
        else:  # mixed effects
            residuals, V, coef, SSE, r, dr = _model_univariate_mixed_effects(self, Y)

    else:  # multivariate
        if n_random_effects > 1:
            raise ValueError("Multivariate mixed effects models not implemented.")
        residuals, V, coef, SSE = _model_multivariate_fixed_effects(self, Y)
    return residuals, V, coef, SSE, r, dr


def _model_univariate_fixed_effects(
    self, Y: np.ndarray
) -> Tuple[np.ndarray, Optional[np.ndarray], np.ndarray, np.ndarray]:
    """Runs a univariate fixed effects linear model

    Parameters
    ----------
    Y : numpy.array
        Response variable matrix

    Returns
    -------
    numpy.array
        Model residuals.
    numpy.array
        Variance matrix bases.
    numpy.array
        Model coefficients.
    numpy.array
        Sum of squared errors.
    """
    if self.V is None:  # OLS
        coef = la.pinv(self.X) @ Y
        residuals = Y - self.X @ coef
        V = None

    else:
        V = self.V / np.diag(self.V).mean(0)
        Vmh = la.inv(la.cholesky(V).T)

        coef = (la.pinv(Vmh @ self.X) @ Vmh) @ Y
        residuals = Vmh @ Y - (Vmh @ self.X) @ coef

    SSE = np.sum(residuals ** 2, axis=0)
    SSE = SSE[None]

    return residuals, V, coef, SSE


def _model_univariate_mixed_effects(self, Y: np.ndarray) -> Tuple[np.ndarray, ...]:
    """Runs a univariate linear mixed effects model.

    Parameters
    ----------
    Y : numpy.array
        Response variable matrix

    Returns
    -------
    numpy.array
        Model residuals.
    numpy.array
        Variance matrix bases.
    numpy.array
        Model coefficients.
    numpy.array
        Sum of squared errors.
    numpy.array
        Matrix of coefficients of the components of slm.V divided by their sum.
    numpy.array
        Vector of increments in slm.r
    """
    n_samples, n_vertices = Y.shape[:2]
    n_random_effects = _get_n_random_effects(self)
    n_predictors = self.X.shape[1]

    q1 = n_random_effects - 1

    V = self.V / np.diagonal(self.V, axis1=0, axis2=1).mean(-1)
    slm_r = np.zeros((q1, n_vertices))

    # start Fisher scoring algorithm
    R = np.eye(n_samples) - self.X @ la.pinv(self.X)
    RVV = (V.T @ R.T).T
    E = Y.T @ (R.T @ RVV.T)
    E *= Y.T
    E = E.sum(-1)

    RVV2 = np.zeros([n_samples, n_samples, n_random_effects])
    E2 = np.zeros([n_random_effects, n_vertices])
    for j in range(n_random_effects):
        RV2 = R @ V[..., j]
        E2[j] = (Y * ((RV2 @ R) @ Y)).sum(0)
        RVV2[..., j] = RV2

    M = np.einsum("ijk,jil->kl", RVV, RVV, optimize="optimal")

    theta = la.pinv(M) @ E
    tlim = np.sqrt(2 * np.diag(la.pinv(M))) * self.thetalim
    tlim = tlim[:, None] * theta.sum(0)
    m = theta < tlim
    theta[m] = tlim[m]
    r = theta[:q1] / theta.sum(0)

    Vt = 2 * la.pinv(M)
    m1 = np.diag(Vt)
    m2 = 2 * Vt.sum(0)
    Vr = m1[:q1] - m2[:q1] * slm_r.mean(1) + Vt.sum() * (r ** 2).mean(-1)
    dr = np.sqrt(Vr) * self.drlim

    # Extra Fisher scoring iterations
    for _ in range(self.niter):
        irs = np.round(r.T / dr)
        ur, jr = np.unique(irs, axis=0, return_inverse=True)
        nr = ur.shape[0]
        for ir in range(nr):
            iv = jr == ir
            rv = r[:, iv].mean(1)

            Vs = (1 - rv.sum()) * V[..., n_random_effects - 1]
            Vs += (V[..., :q1] * rv).sum(-1)

            Vinv = la.inv(Vs)
            VinvX = Vinv @ self.X
            G = la.pinv(self.X.T @ VinvX) @ VinvX.T
            R = Vinv - VinvX @ G

            RVV = (V.T @ R.T).T
            E = Y[:, iv].T @ (R.T @ RVV.T)
            E *= Y[:, iv].T
            E = E.sum(-1)

            M = np.einsum("ijk,jil->kl", RVV, RVV, optimize="optimal")

            thetav = la.pinv(M) @ E
            tlim = np.sqrt(2 * np.diag(la.pinv(M))) * self.thetalim
            tlim = tlim[:, None] * thetav.sum(0)

            m = thetav < tlim
            thetav[m] = tlim[m]
            theta[:, iv] = thetav

        r = theta[:q1] / theta.sum(0)

    # finish Fisher scoring
    irs = np.round(r.T / dr)
    ur, jr = np.unique(irs, axis=0, return_inverse=True)
    nr = ur.shape[0]

    coef = np.zeros((n_predictors, n_vertices))
    SSE = np.zeros(n_vertices)
    residuals = Y
    for ir in range(nr):
        iv = jr == ir
        rv = r[:, iv].mean(1)

        Vs = (1 - rv.sum()) * V[..., n_random_effects - 1]
        Vs += (V[..., :q1] * rv).sum(-1)

        # Vmh = la.inv(la.cholesky(Vs).T)
        Vmh = la.inv(la.cholesky(Vs))
        VmhX = Vmh @ self.X
        G = (la.pinv(VmhX.T @ VmhX) @ VmhX.T) @ Vmh

        coef[:, iv] = G @ residuals[:, iv]
        R = Vmh - VmhX @ G
        residuals[:, iv] = R @ residuals[:, iv]
        SSE[iv] = (residuals[:, iv] ** 2).sum(0)
    SSE = SSE[None]
    return residuals, V, coef, SSE, r, dr[:, None]


def _model_multivariate_fixed_effects(
    self, Y: np.ndarray
) -> Tuple[np.ndarray, Optional[np.ndarray], np.ndarray, np.ndarray]:
    """Runs a multivariate linear fixed effects model.

    Parameters
    ----------
    Y : numpy.array
        Response variable matrix

    Returns
    -------
    numpy.array
        Model residuals.
    numpy.array
        Variance matrix bases.
    numpy.array
        Model coefficients.
    numpy.array
        Sum of squared errors.
    """
    k = Y.shape[2]
    n_vertices = Y.shape[1]

    if self.V is None:
        X2 = self.X
        V = self.V
        pinvX = la.pinv(self.X)
    else:
        V = self.V / np.diag(self.V).mean(0)
        Vmh = la.inv(la.cholesky(V)).T
        X2 = Vmh @ self.X
        pinvX = la.pinv(X2)
        Y = Vmh @ Y

    coef = pinvX @ Y.T.swapaxes(-1, -2)
    residuals = Y - (X2 @ coef).swapaxes(-1, -2).T
    coef = coef.swapaxes(-1, -2).T

    k2 = k * (k + 1) // 2
    SSE = np.zeros((k2, n_vertices))
    j = -1
    for j1 in range(k):
        for j2 in range(j1 + 1):
            j = j + 1
            SSE[j] = (residuals[..., j1] * residuals[..., j2]).sum(0)
    return residuals, V, coef, SSE


def _get_design_matrix(self, n_samples: int) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """Wrapper for fetching the design matrix

    Parameters
    ----------
    n_samples: int
        Scalar describing the number of samples.

    Returns
    -------
    numpy.array
        Design matrix
    numpy.array, None
        Variance matrix bases. Returns None for fixed effects models.

    """
    if isinstance(self.model, MixedEffect):
        X, V = _get_mixed_design(self)
    else:
        X = _get_fixed_design(self, n_samples)
        V = None
    return X, V


def _get_mixed_design(self) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """Fetches the design matrix from a mixed effects model.

    Parameters
    ----------
    n_samples : numpy.array
        Scalar describing the number of samples.

    Returns
    -------
    numpy.array
        Design matrix
    numpy.array
        Variance matrix bases.
    """
    X = self.model.mean.matrix.values
    n_samples = X.shape[0]
    random_effects = self.model.variance.matrix.values

    # check in var contains intercept (constant term)
    _, n_random_effects = random_effects.shape
    identity = np.identity(n_samples).ravel()

    r = identity - random_effects @ (la.pinv(random_effects) @ identity)
    if (r ** 2).mean() > np.finfo(float).eps:
        warnings.warn("Did you forget an error term, I? :-)")

    if (
        n_random_effects > 1
        or n_random_effects == 1
        and np.abs(identity - random_effects.T).sum() > 0
    ):
        V = random_effects.reshape(n_samples, n_samples, -1)
    else:
        V = None

    return X, V


def _get_fixed_design(self, n_samples: int) -> np.ndarray:
    """Fetches the design matrix from a fixed effects model.

    Parameters
    ----------
    n_samples : int
        Scalar describing the number of samples.

    Returns
    -------
    numpy.array
        Design matrix
    """
    if isinstance(self.model, FixedEffect):
        X = self.model.matrix.values
    else:
        if self.model.size > 1:
            warnings.warn(
                "If you don"
                "'t convert vectors to terms you can "
                "get unexpected results :-("
            )
        X = self.model

    if X.shape[0] == 1:
        X = np.tile(X, (n_samples, 1))
    return X


def _compute_resls(self, Y: np.ndarray) -> Tuple[np.ndarray, dict]:
    """Computes the sum over observations of squares of differences of
    normalized residuals along each edge.

    Parameters
    ----------
    Y : numpy.array
        Response variable residual matrix.

    Returns
    -------
    numpy.array
        Sum over observations of squares of differences of normalized residuals
        along each edge.
    dict
        Dictionary containing the mesh connections in either triangle or lattice
        format. The dictionary's sole key is 'tri' for triangle connections or
        'lat' for lattice connections.
    """
    if isinstance(self.surf, BSPolyData):
        mesh_connections = {"tri": np.array(get_cells(self.surf)) + 1}
    else:
        key = "tri" if "tri" in self.surf else "lat"
        mesh_connections = {key: self.surf[key]}

    edges = mesh_edges(self.surf, self.mask)

    n_edges = edges.shape[0]

    Y = np.atleast_3d(Y)
    resl = np.zeros((n_edges, Y.shape[2]))

    for j in range(Y.shape[2]):
        normr = np.sqrt(self.SSE[((j + 1) * (j + 2) // 2) - 1])
        for i in range(Y.shape[0]):
            u = Y[i, :, j] / normr
            resl[:, j] += np.diff(u[edges], axis=1).ravel() ** 2

    return resl, mesh_connections


def _check_constant_term(X: np.ndarray) -> None:
    """Checks whether an error term was provided.

    Parameters
    ----------
    X : numpy.array
        Design matrix.
    """
    r = 1 - X @ la.pinv(X).sum(1)
    if (r ** 2).mean() > np.finfo(float).eps:
        warnings.warn("Did you forget an error term, I? :-)")


def _get_n_random_effects(self) -> int:
    """Gets the number of random effects.

    Returns
    -------
    int
        Number of random effects.
    """
    if isinstance(self.model, MixedEffect):
        n_random_effects = self.model.variance.matrix.values.shape[1]
    else:
        n_random_effects = 1
    return n_random_effects



"""Data generation for peak_clus unit tests."""

import pickle

import numpy as np
from sklearn.model_selection import ParameterGrid

from brainstat.mesh.utils import mesh_edges
from brainstat.stats._multiple_comparisons import peak_clus
from brainstat.tests.testutil import datadir, generate_slm


def generate_random_slm(I):
    slm = generate_slm(
        t=I["t"],
        df=I["df"],
        k=I["k"],
        resl=I["resl"],
        tri=I["tri"],
        surf=I,
        mask=I["mask"],
        cluster_threshold=I["thresh"],
    )
    return slm


def generate_peak_clus_out(slm, I):
    D = {}
    D["peak"], D["clus"], D["clusid"] = peak_clus(
        slm, I["thresh"], reselspvert=I["reselspvert"], edg=I["edg"]
    )
    return D


def params2files(I, D, test_num):
    """Converts params to input/output files"""
    basename = "xstatpeakc"
    fin_name = datadir(basename + "_" + f"{test_num:02d}" + "_IN.pkl")
    fout_name = datadir(basename + "_" + f"{test_num:02d}" + "_OUT.pkl")
    with open(fin_name, "wb") as f:
        pickle.dump(I, f, protocol=4)
    with open(fout_name, "wb") as g:
        pickle.dump(D, g, protocol=4)
    return


def generate_test_data():
    np.random.seed(0)

    # generate the parameters
    tri = np.random.randint(1, int(50), size=(100, 3))
    coord = np.random.rand(3, 50)
    edg = mesh_edges({"tri": tri})
    n_edges = edg.shape[0]
    n_vertices = int(tri.shape[0])
    cluster_threshold = np.random.rand()
    mygrid = [
        {
            "num_t": [1, 2, 3],
            "k": [1, 2, 3],
            "df": [1, [1, 1]],
            "mask": [False, True],
            "reselspvert": [None, True],
        },
    ]
    myparamgrid = ParameterGrid(mygrid)

    # Generate data.
    test_num = 0
    for params in myparamgrid:
        I = {
            "tri": tri,
            "edg": edg,
            "thresh": cluster_threshold,
            "t": np.random.random_sample((params["num_t"], n_vertices)),
            "resl": np.random.random_sample((n_edges, 1)),
            "k": params["k"],
            "df": params["df"],
            "coord": coord,
        }

        if params["mask"] is True:
            I["mask"] = np.random.choice(a=[False, True], size=(n_vertices))
        else:
            I["mask"] = np.ones((n_vertices), dtype=bool)

        if params["reselspvert"] is True:
            I["reselspvert"] = np.random.rand(n_vertices)
        else:
            I["reselspvert"] = None

        # Here we go: generate slm & run peak_clus & save in-out
        slm = generate_random_slm(I)
        D = generate_peak_clus_out(slm, I)
        test_num += 1
        params2files(I, D, test_num)


if __name__ == "__main__":
    generate_test_data()



import math
import sys
import warnings

import numpy as np
from scipy.linalg import cholesky, null_space

from .terms import FixedEffect


def _t_test(self) -> None:
    """T statistics for a contrast in a univariate or multivariate model.

    Parameters
    ----------
    self : brainstat.stats.SLM.SLM
        SLM object that has already run linear_model


    """

    if isinstance(self.contrast, FixedEffect):
        self.contrast = self.contrast.m.to_numpy()

    if not isinstance(self.df, np.ndarray):
        self.df = np.array([self.df])

    if self.contrast.ndim == 1:
        self.contrast = np.reshape(self.contrast, (-1, 1))

    [n, p] = np.shape(self.X)
    pinvX = np.linalg.pinv(self.X)

    if len(self.contrast) <= p:
        c = np.concatenate(
            (self.contrast, np.zeros((1, p - np.shape(self.contrast)[1]))), axis=1
        ).T

        if np.square(np.dot(null_space(self.X).T, c)).sum() / np.square(
            c
        ).sum() > np.spacing(1):
            sys.exit("Contrast is not estimable :-(")

    else:
        c = np.dot(pinvX, self.contrast)
        r = self.contrast - np.dot(self.X, c)

        if np.square(np.ravel(r, "F")).sum() / np.square(
            np.ravel(self.contrast, "F")
        ).sum() > np.spacing(1):
            warnings.warn("Contrast is not in the model :-( ")

    self.c = c.T
    self.df = self.df[len(self.df) - 1]

    if np.ndim(self.coef) == 2:
        k = 1
        self.k = k

        if self.r is None:
            # fixed effect
            if self.V is not None:
                Vmh = np.linalg.inv(cholesky(self.V).T)
                pinvX = np.linalg.pinv(np.dot(Vmh, self.X))
            Vc = np.sum(np.square(np.dot(c.T, pinvX)), axis=1)
        else:
            # mixed effect
            q1, v = np.shape(self.r)
            q = q1 + 1
            nc = np.shape(self.dr)[1]
            chunk = math.ceil(v / nc)
            irs = np.zeros((q1, v))

            for ic in range(1, nc + 1):
                v1 = 1 + (ic - 1) * chunk
                v2 = np.min((v1 + chunk - 1, v))

                vc = v2 - v1 + 1

                irs[:, int(v1 - 1) : int(v2)] = np.around(
                    np.multiply(
                        self.r[:, int(v1 - 1) : int(v2)],
                        np.tile(1 / self.dr[:, (ic - 1)], (1, vc)),
                    )
                )

            ur, _, jr = np.unique(irs.T, axis=0, return_index=True, return_inverse=True)
            jr = jr + 1
            nr = np.shape(ur)[0]
            self.dfs = np.zeros((1, v))
            Vc = np.zeros((1, v))

            for ir in range(1, nr + 1):
                iv = (jr == ir).astype(bool)
                rv = self.r[:, iv].mean(axis=1)
                V = (1 - rv.sum()) * self.V[:, :, (q - 1)]

                for j in range(1, q1 + 1):
                    V = V + rv[(j - 1)] * self.V[:, :, (j - 1)]

                Vinv = np.linalg.inv(V)
                VinvX = np.dot(Vinv, self.X)
                Vbeta = np.linalg.pinv(np.dot(self.X.T, VinvX))
                G = np.dot(Vbeta, VinvX.T)
                Gc = np.dot(G.T, c)
                R = Vinv - np.dot(VinvX, G)
                E = np.zeros((q, 1))
                RVV = np.zeros((np.shape(self.V)))
                M = np.zeros((q, q))

                for j in range(1, q + 1):
                    E[(j - 1)] = np.dot(Gc.T, np.dot(self.V[:, :, (j - 1)], Gc))
                    RVV[:, :, (j - 1)] = np.dot(R, self.V[:, :, (j - 1)])

                for j1 in range(1, q + 1):
                    for j2 in range(j1, q + 1):
                        M[(j1 - 1), (j2 - 1)] = (
                            RVV[:, :, (j1 - 1)] * RVV[:, :, (j2 - 1)].T
                        ).sum()
                        M[(j2 - 1), (j1 - 1)] = M[(j1 - 1), (j2 - 1)]

                vc = np.dot(c.T, np.dot(Vbeta, c))
                Vc[0, iv] = vc
                self.dfs[0, iv] = np.square(vc) / np.dot(
                    E.T, np.dot(np.linalg.pinv(M), E)
                )

        self.ef = np.dot(c.T, self.coef)
        self.sd = np.sqrt(np.multiply(Vc, self.SSE) / self.df)
        self.t = np.multiply(
            np.divide(self.ef, (self.sd + (self.sd <= 0))), self.sd > 0
        )

    else:
        # multivariate
        p, v, k = np.shape(self.coef)
        self.k = k
        self.ef = np.zeros((k, v))

        for j in range(0, k):
            self.ef[j, :] = np.dot(c.T, self.coef[:, :, j])

        j = np.arange(1, k + 1)
        jj = (np.multiply(j, j + 1) / 2) - 1
        jj = jj.astype(int)

        vf = np.divide(np.sum(np.square(np.dot(c.T, pinvX)), axis=1), self.df)
        self.sd = np.sqrt(vf * self.SSE[jj, :])

        if k == 2:
            det = np.multiply(self.SSE[0, :], self.SSE[2, :]) - np.square(
                self.SSE[1, :]
            )

            self.t = (
                np.multiply(np.square(self.ef[0, :]), self.SSE[2, :])
                + np.multiply(np.square(self.ef[1, :]), self.SSE[0, :])
                - np.multiply(
                    np.multiply(2 * self.ef[0, :], self.ef[1, :]), self.SSE[1, :]
                )
            )

        if k == 3:
            det = (
                np.multiply(
                    self.SSE[0, :],
                    (
                        np.multiply(self.SSE[2, :], self.SSE[5, :])
                        - np.square(self.SSE[4, :])
                    ),
                )
                - np.multiply(self.SSE[5, :], np.square(self.SSE[1, :]))
                + np.multiply(
                    self.SSE[3, :],
                    (
                        np.multiply(self.SSE[1, :], self.SSE[4, :]) * 2
                        - np.multiply(self.SSE[2, :], self.SSE[3, :])
                    ),
                )
            )

            self.t = np.multiply(
                np.square(self.ef[0, :]),
                (
                    np.multiply(self.SSE[2, :], self.SSE[5, :])
                    - np.square(self.SSE[4, :])
                ),
            )

            self.t = self.t + np.multiply(
                np.square(self.ef[1, :]),
                (
                    np.multiply(self.SSE[0, :], self.SSE[5, :])
                    - np.square(self.SSE[3, :])
                ),
            )

            self.t = self.t + np.multiply(
                np.square(self.ef[2, :]),
                (
                    np.multiply(self.SSE[0, :], self.SSE[2, :])
                    - np.square(self.SSE[1, :])
                ),
            )

            self.t = self.t + np.multiply(
                2 * self.ef[0, :],
                np.multiply(
                    self.ef[1, :],
                    (
                        np.multiply(self.SSE[3, :], self.SSE[4, :])
                        - np.multiply(self.SSE[1, :], self.SSE[5, :])
                    ),
                ),
            )

            self.t = self.t + np.multiply(
                2 * self.ef[0, :],
                np.multiply(
                    self.ef[2, :],
                    (
                        np.multiply(self.SSE[1, :], self.SSE[4, :])
                        - np.multiply(self.SSE[2, :], self.SSE[3, :])
                    ),
                ),
            )

            self.t = self.t + np.multiply(
                2 * self.ef[1, :],
                np.multiply(
                    self.ef[2, :],
                    (
                        np.multiply(self.SSE[1, :], self.SSE[3, :])
                        - np.multiply(self.SSE[0, :], self.SSE[4, :])
                    ),
                ),
            )

        if k > 3:
            sys.exit("Hotelling" "s T for k>3 not programmed yet")

        self.t = np.multiply(np.divide(self.t, (det + (det <= 0))), (det > 0)) / vf
        self.t = np.multiply(np.sqrt(self.t + (self.t <= 0)), (self.t > 0))
    self.t = np.atleast_2d(self.t)



"""Configurations for pytest."""
import git


# This hook will run by pytest one time before execution all the particular.
# tests. It does checkout the needed 'extern/test-data' directory from the
# separate branch 'test-data'.
def pytest_configure(config):
    repo = git.Repo(search_parent_directories=True)
    origin = repo.remote()

    cli = origin.repo.git
    cli.checkout("origin/test-data-2.0", "--", "extern/test-data")
    # we have to reset the test-data, as checkout is staging it to the git index
    cli.reset("--", "extern/test-data")



"""Unit tests of stat_threshold."""
import pickle

import numpy as np

from brainstat.stats._multiple_comparisons import stat_threshold

from .testutil import datadir


def dummy_test(infile, expfile):

    # load input test data
    ifile = open(infile, "br")
    idic = pickle.load(ifile)
    ifile.close()

    # run stat_xthreshold
    A, B, C, D, E, F = stat_threshold(
        idic["search_volume"],
        idic["num_voxels"],
        idic["fwhm"],
        idic["df"],
        idic["p_val_peak"],
        idic["cluster_threshold"],
        idic["p_val_extent"],
        idic["nconj"],
        idic["nvar"],
        None,
        idic["nprint"],
    )
    outdic = {
        "peak_threshold": A,
        "extent_threshold": B,
        "peak_threshold_1": C,
        "extent_threshold_1": D,
        "t": E,
        "rho": F,
    }

    # load expected outout data
    efile = open(expfile, "br")
    expdic = pickle.load(efile)
    efile.close()

    testout = []

    for key in outdic.keys():
        comp = np.allclose(outdic[key], expdic[key], rtol=1e-05, equal_nan=True)
        testout.append(comp)

    assert all(flag == True for (flag) in testout)


# parameters in *pck is equal to default params, if not specified in tests


def test_01():
    infile = datadir("xthresh_01_IN.pkl")
    expfile = datadir("xthresh_01_OUT.pkl")
    dummy_test(infile, expfile)


def test_02():
    infile = datadir("xthresh_02_IN.pkl")
    expfile = datadir("xthresh_02_OUT.pkl")
    dummy_test(infile, expfile)


def test_03():
    infile = datadir("xthresh_03_IN.pkl")
    expfile = datadir("xthresh_03_OUT.pkl")
    dummy_test(infile, expfile)


def test_04():
    infile = datadir("xthresh_04_IN.pkl")
    expfile = datadir("xthresh_04_OUT.pkl")
    dummy_test(infile, expfile)


def test_05():
    infile = datadir("xthresh_05_IN.pkl")
    expfile = datadir("xthresh_05_OUT.pkl")
    dummy_test(infile, expfile)


def test_06():
    infile = datadir("xthresh_06_IN.pkl")
    expfile = datadir("xthresh_06_OUT.pkl")
    dummy_test(infile, expfile)


def test_07():
    infile = datadir("xthresh_07_IN.pkl")
    expfile = datadir("xthresh_07_OUT.pkl")
    dummy_test(infile, expfile)


def test_08():
    infile = datadir("xthresh_08_IN.pkl")
    expfile = datadir("xthresh_08_OUT.pkl")
    dummy_test(infile, expfile)


def test_09():
    infile = datadir("xthresh_09_IN.pkl")
    expfile = datadir("xthresh_09_OUT.pkl")
    dummy_test(infile, expfile)


def test_10():
    infile = datadir("xthresh_10_IN.pkl")
    expfile = datadir("xthresh_10_OUT.pkl")
    dummy_test(infile, expfile)


def test_11():
    infile = datadir("xthresh_11_IN.pkl")
    expfile = datadir("xthresh_11_OUT.pkl")
    dummy_test(infile, expfile)


def test_12():
    infile = datadir("xthresh_12_IN.pkl")
    expfile = datadir("xthresh_12_OUT.pkl")
    dummy_test(infile, expfile)


def test_13():
    infile = datadir("xthresh_13_IN.pkl")
    expfile = datadir("xthresh_13_OUT.pkl")
    dummy_test(infile, expfile)


def test_14():
    infile = datadir("xthresh_14_IN.pkl")
    expfile = datadir("xthresh_14_OUT.pkl")
    dummy_test(infile, expfile)


def test_15():
    infile = datadir("xthresh_15_IN.pkl")
    expfile = datadir("xthresh_15_OUT.pkl")
    dummy_test(infile, expfile)


def test_16():
    infile = datadir("xthresh_16_IN.pkl")
    expfile = datadir("xthresh_16_OUT.pkl")
    dummy_test(infile, expfile)


def test_17():
    infile = datadir("xthresh_17_IN.pkl")
    expfile = datadir("xthresh_17_OUT.pkl")
    dummy_test(infile, expfile)


def test_18():
    infile = datadir("xthresh_18_IN.pkl")
    expfile = datadir("xthresh_18_OUT.pkl")
    dummy_test(infile, expfile)


def test_19():
    infile = datadir("xthresh_19_IN.pkl")
    expfile = datadir("xthresh_19_OUT.pkl")
    dummy_test(infile, expfile)


def test_20():
    infile = datadir("xthresh_20_IN.pkl")
    expfile = datadir("xthresh_20_OUT.pkl")
    dummy_test(infile, expfile)


def test_21():
    infile = datadir("xthresh_21_IN.pkl")
    expfile = datadir("xthresh_21_OUT.pkl")
    dummy_test(infile, expfile)


def test_22():
    infile = datadir("xthresh_22_IN.pkl")
    expfile = datadir("xthresh_22_OUT.pkl")
    dummy_test(infile, expfile)


def test_23():
    infile = datadir("xthresh_23_IN.pkl")
    expfile = datadir("xthresh_23_OUT.pkl")
    dummy_test(infile, expfile)


def test_24():
    infile = datadir("xthresh_24_IN.pkl")
    expfile = datadir("xthresh_24_OUT.pkl")
    dummy_test(infile, expfile)


def test_25():
    infile = datadir("xthresh_25_IN.pkl")
    expfile = datadir("xthresh_25_OUT.pkl")
    dummy_test(infile, expfile)


def test_26():
    infile = datadir("xthresh_26_IN.pkl")
    expfile = datadir("xthresh_26_OUT.pkl")
    dummy_test(infile, expfile)


def test_27():
    infile = datadir("xthresh_27_IN.pkl")
    expfile = datadir("xthresh_27_OUT.pkl")
    dummy_test(infile, expfile)


def test_28():
    infile = datadir("xthresh_28_IN.pkl")
    expfile = datadir("xthresh_28_OUT.pkl")
    dummy_test(infile, expfile)


def test_29():
    infile = datadir("xthresh_29_IN.pkl")
    expfile = datadir("xthresh_29_OUT.pkl")
    dummy_test(infile, expfile)


def test_30():
    infile = datadir("xthresh_30_IN.pkl")
    expfile = datadir("xthresh_30_OUT.pkl")
    dummy_test(infile, expfile)


def test_31():
    infile = datadir("xthresh_31_IN.pkl")
    expfile = datadir("xthresh_31_OUT.pkl")
    dummy_test(infile, expfile)


def test_32():
    infile = datadir("xthresh_32_IN.pkl")
    expfile = datadir("xthresh_32_OUT.pkl")
    dummy_test(infile, expfile)


def test_33():
    infile = datadir("xthresh_33_IN.pkl")
    expfile = datadir("xthresh_33_OUT.pkl")
    dummy_test(infile, expfile)


def test_34():
    infile = datadir("xthresh_34_IN.pkl")
    expfile = datadir("xthresh_34_OUT.pkl")
    dummy_test(infile, expfile)


def test_35():
    infile = datadir("xthresh_35_IN.pkl")
    expfile = datadir("xthresh_35_OUT.pkl")
    dummy_test(infile, expfile)


def test_36():
    infile = datadir("xthresh_36_IN.pkl")
    expfile = datadir("xthresh_36_OUT.pkl")
    dummy_test(infile, expfile)


def test_37():
    infile = datadir("xthresh_37_IN.pkl")
    expfile = datadir("xthresh_37_OUT.pkl")
    dummy_test(infile, expfile)


def test_38():
    infile = datadir("xthresh_38_IN.pkl")
    expfile = datadir("xthresh_38_OUT.pkl")
    dummy_test(infile, expfile)


def test_39():
    infile = datadir("xthresh_39_IN.pkl")
    expfile = datadir("xthresh_39_OUT.pkl")
    dummy_test(infile, expfile)


def test_40():
    infile = datadir("xthresh_40_IN.pkl")
    expfile = datadir("xthresh_40_OUT.pkl")
    dummy_test(infile, expfile)


def test_41():
    infile = datadir("xthresh_41_IN.pkl")
    expfile = datadir("xthresh_41_OUT.pkl")
    dummy_test(infile, expfile)


def test_42():
    infile = datadir("xthresh_42_IN.pkl")
    expfile = datadir("xthresh_42_OUT.pkl")
    dummy_test(infile, expfile)


def test_43():
    infile = datadir("xthresh_43_IN.pkl")
    expfile = datadir("xthresh_43_OUT.pkl")
    dummy_test(infile, expfile)


def test_44():
    infile = datadir("xthresh_44_IN.pkl")
    expfile = datadir("xthresh_44_OUT.pkl")
    dummy_test(infile, expfile)


def test_45():
    infile = datadir("xthresh_45_IN.pkl")
    expfile = datadir("xthresh_45_OUT.pkl")
    dummy_test(infile, expfile)


def test_46():
    infile = datadir("xthresh_46_IN.pkl")
    expfile = datadir("xthresh_46_OUT.pkl")
    dummy_test(infile, expfile)


def test_47():
    infile = datadir("xthresh_47_IN.pkl")
    expfile = datadir("xthresh_47_OUT.pkl")
    dummy_test(infile, expfile)


def test_48():
    infile = datadir("xthresh_48_IN.pkl")
    expfile = datadir("xthresh_48_OUT.pkl")
    dummy_test(infile, expfile)



""" Histology context decoder """
from pathlib import Path
from typing import Callable, Optional, Union

import h5py
import numpy as np
from brainspace.gradient.gradient import GradientMaps
from brainspace.utils.parcellation import reduce_by_labels

from brainstat._typing import ArrayLike
from brainstat._utils import (
    _download_file,
    data_directories,
    logger,
    read_data_fetcher_json,
)


def compute_histology_gradients(
    mpc: np.ndarray,
    kernel: Union[str, Callable[[np.ndarray], np.ndarray]] = "normalized_angle",
    approach: Union[str, Callable[[np.ndarray], np.ndarray]] = "dm",
    n_components: int = 10,
    alignment: Optional[str] = None,
    random_state: Optional[int] = None,
    gamma: Optional[float] = None,
    sparsity: Optional[float] = 0.9,
    reference: Optional[np.ndarray] = None,
    n_iter: int = 10,
) -> GradientMaps:
    """Computes microstructural profile covariance gradients.

    Parameters
    ----------
    mpc : numpy.ndarray
        Microstructural profile covariance matrix.
    kernel : str, optional
         Kernel function to build the affinity matrix. Possible options: {‘pearson’,
         ‘spearman’, ‘cosine’, ‘normalized_angle’, ‘gaussian’}. If callable, must
         receive a 2D array and return a 2D square array. If None, use input matrix.
         By default "normalized_angle".
    approach : str, optional
        Embedding approach. Can be 'pca' for Principal Component Analysis, 'le' for
        laplacian eigenmaps, or 'dm' for diffusion mapping, by default "dm".
    n_components : int, optional
        Number of components to return, by default 10.
    alignment : str, None, optional
        Alignment approach. Only used when two or more datasets are provided.
        Valid options are 'pa' for procrustes analysis and "joint" for joint embedding.
        If None, no alignment is peformed, by default None.
    random_state : int, None, optional
        Random state, by default None
    gamma : float, None, optional
        Inverse kernel width. Only used if kernel == "gaussian". If None, gamma=1/n_feat,
        by default None.
    sparsity : float, optional
        Proportion of smallest elements to zero-out for each row, by default 0.9.
    reference : numpy.ndarray, optional
        Initial reference for procrustes alignments. Only used when
        alignment == 'procrustes', by default None.
    n_iter : int, optional
        Number of iterations for Procrustes alignment, by default 10.

    Returns
    -------
    brainspace.gradient.gradient.GradientMaps
        BrainSpace gradient maps object.
    """
    gm = GradientMaps(
        kernel=kernel,
        approach=approach,
        n_components=n_components,
        alignment=alignment,
        random_state=random_state,
    )
    gm.fit(mpc, gamma=gamma, sparsity=sparsity, n_iter=n_iter, reference=reference)
    return gm


def compute_mpc(profile: np.ndarray, labels: np.ndarray) -> np.ndarray:
    """Computes MPC for given labels on a surface template.

    Parameters
    ----------
    profile : numpy.ndarray
        Histological profiles of size surface-by-vertex.
    labels : numpy.ndarray
        Labels of regions of interest. Use 0 to denote regions that will not be included.

    Returns
    -------
    numpy.ndarray
        Microstructural profile covariance.
    """

    roi_profile = reduce_by_labels(profile, labels)
    if np.any(labels == 0):
        # Remove 0's in the labels.
        roi_profile = roi_profile[:, 1:]

    p_corr = partial_correlation(roi_profile, np.mean(profile, axis=1))

    mpc = 0.5 * np.log((1 + p_corr) / (1 - p_corr))
    mpc[p_corr > 0.99999] = 0  # Deals with floating point issues where p_corr==1
    mpc[mpc == np.inf] = 0
    mpc[mpc == np.nan] = 0

    return mpc


def read_histology_profile(
    data_dir: Optional[Union[str, Path]] = None,
    template: str = "fsaverage",
    overwrite: bool = False,
) -> np.ndarray:
    """Reads BigBrain histology profiles.

    Parameters
    ----------
    data_dir : str, pathlib.Path, None, optional
        Path to the data directory. If data is not found here then data will be
        downloaded. If None, data_dir is set to the home directory, by default None.
    template : str, optional
        Surface template. Currently allowed options are 'fsaverage' and 'fslr32k', by
        default 'fsaverage'.
    overwrite : bool, optional
        If true, existing data will be overwrriten, by default False.

    Returns
    -------
    numpy.ndarray
        Depth-by-vertex array of BigBrain intensities.
    """

    data_dir = Path(data_dir) if data_dir else data_directories["BIGBRAIN_DATA_DIR"]
    histology_file = data_dir / ("histology_" + template + ".h5")

    if not histology_file.exists() or overwrite:
        logger.info(
            "Could not find a histological profile or an overwrite was requested. Downloading..."
        )
        download_histology_profiles(
            data_dir=data_dir, template=template, overwrite=overwrite
        )

    with h5py.File(histology_file, "r") as h5_file:
        return h5_file.get(template)[...]


def download_histology_profiles(
    data_dir: Optional[Union[str, Path]] = None,
    template: str = "fsaverage",
    overwrite: bool = False,
) -> None:
    """Downloads BigBrain histology profiles.

    Parameters
    ----------
    data_dir : str, pathlib,Path, None, optional
        Path to the directory to store the data. If None, defaults to the home
        directory, by default None.
    template : str, optional
        Surface template. Currently allowed options are 'fsaverage' and 'fslr32k', by
        default 'fsaverage'.
    overwrite : bool, optional
        If true, existing data will be overwrriten, by default False.

    Raises
    ------
    KeyError
        Thrown if an invalid template is requested.
    """

    data_dir = Path(data_dir) if data_dir else data_directories["BIGBRAIN_DATA_DIR"]
    data_dir.mkdir(parents=True, exist_ok=True)
    output_file = data_dir / ("histology_" + template + ".h5")

    url = read_data_fetcher_json()["bigbrain_profiles"][template]["url"]

    try:
        _download_file(url, output_file, overwrite)
    except KeyError:
        raise KeyError(
            "Could not find the requested template. Valid templates are: 'fslr32k', 'fsaverage', 'fsaverage5'."
        )


def partial_correlation(X: ArrayLike, covar: np.ndarray) -> np.ndarray:
    """Runs a partial correlation whilst correcting for a covariate.

    Parameters
    ----------
    X : ArrayLike
        Two-dimensional array of the data to be correlated.
    covar : numpy.ndarray
        One-dimensional array of the covariate.

    Returns
    -------
    numpy.ndarray
        Partial correlation matrix.
    """
    X_mean = np.concatenate((X, covar[:, None]), axis=1)
    pearson_correlation = np.corrcoef(X_mean, rowvar=False)
    r_xy = pearson_correlation[:-1, :-1]
    r_xz = pearson_correlation[0:-1, -1][:, None]

    return (r_xy - r_xz @ r_xz.T) / (np.sqrt(1 - r_xz ** 2) * np.sqrt(1 - r_xz.T ** 2))



import sys
from typing import Any

""" Python 3.6/3.7 compatibility for the ArrayLike type hint. 
In these versions we simply accept anything. The tests in 3.8+
should catch all errors anyway. """
if sys.version_info[1] >= 8:
    from numpy.typing import ArrayLike
else:
    ArrayLike = Any



"""Data generation for SLM unit tests."""

import os
import pickle

import nibabel as nib
import numpy as np
from brainspace.mesh.mesh_elements import get_cells, get_points
from nilearn.datasets import fetch_surf_fsaverage
from sklearn.model_selection import ParameterGrid

import brainstat
from brainstat.context.utils import read_surface_gz
from brainstat.stats.SLM import SLM
from brainstat.stats.terms import FixedEffect, MixedEffect
from brainstat.tests.testutil import datadir, slm2dict
from brainstat.tutorial.utils import fetch_tutorial_data


def slm2files(slm, basename, test_num, input=True):
    """Converts an SLM to its input/output files

    Parameters
    ----------
    slm : brainstat.stats.SLM
        SLM object.
    basename : str
        Base name for the file.
    test_num : int
        Number of the test.
    """
    D = slm2dict(slm)
    D["model"] = "fixed" if type(D["model"]) is FixedEffect else "mixed"
    dict2pkl(D, basename, test_num, input)


def dict2pkl(D, basename, test_num, input=True):
    if "surf" in D and D["surf"] is not None:
        D["surf"] = {
            "tri": np.array(get_cells(D["surf"])),
            "coord": np.array(get_points(D["surf"])).T,
        }

    if "_tri" in D:
        D.pop("_tri")

    if "_surf" in D and D["_surf"] is not None:
        D["surf"] = {
            "tri": np.array(get_cells(D["_surf"])),
            "coord": np.array(get_points(D["_surf"])),
        }
        D.pop("_surf")

    if input:
        stage = "IN"
    else:
        stage = "OUT"
    filename = datadir(basename + "_" + f"{test_num:02d}" + "_" + stage + ".pkl")
    with open(filename, "wb") as f:
        pickle.dump(D, f, protocol=4)


def load_training_data(n):
    brainstat_dir = os.path.dirname(brainstat.__file__)
    data_dir = os.path.join(brainstat_dir, "tutorial")

    tutorial_data = fetch_tutorial_data(n_subjects=n, data_dir=data_dir)
    age = tutorial_data["demographics"]["AGE"].to_numpy()
    iq = tutorial_data["demographics"]["IQ"].to_numpy()

    # Reshape the thickness files such that left and right hemispheres are in the same row.
    files = np.reshape(np.array(tutorial_data["image_files"]), (-1, 2))

    thickness = np.zeros((n, 10242))
    for i in range(n):
        thickness[i, :] = np.squeeze(nib.load(files[i, 0]).get_fdata())
    mask = np.all(thickness != 0, axis=0)

    pial = read_surface_gz(fetch_surf_fsaverage()["pial_left"])
    return (pial, mask, age, iq, thickness)


def generate_test_data():
    pial, mask, age, iq, thickness = load_training_data(n=20)
    fixed_model = FixedEffect(1) + FixedEffect(age, "age")
    mixed_model = (
        FixedEffect(1)
        + FixedEffect(age, "age")
        + MixedEffect(iq, name_ran="iq")
        + MixedEffect(1, name_ran="Identity")
    )

    variates_2 = np.concatenate(
        (thickness[:, :, None], np.random.random_sample(thickness.shape)[:, :, None]),
        axis=2,
    )
    variates_3 = np.concatenate(
        (
            thickness[:, :, None],
            np.random.rand(thickness.shape[0], thickness.shape[1], 2),
        ),
        axis=2,
    )

    # Params 1: No surface, fixed effect.
    # Params 2: One-tailed mixed with theta/dr changes.
    # Params 3: With surface. and RFT correction.
    parameters = [
        {
            "Y": [thickness, variates_2, variates_3],
            "model": [fixed_model],
            "contrast": [-age],
            "correction": [None, "fdr"],
            "surf": [None],
            "mask": [mask],
            "niter": [1],
            "thetalim": [0.01],
            "drlim": [0.1],
            "two_tailed": [True],
            "cluster_threshold": [0.001],
        },
        {
            "Y": [thickness],
            "model": [mixed_model],
            "contrast": [-age],
            "correction": ["fdr"],
            "surf": [None, pial],
            "mask": [mask],
            "niter": [1],
            "thetalim": [0.01, 0.05],
            "drlim": [0.1, 0.2],
            "two_tailed": [False],
            "cluster_threshold": [0.001],
        },
        {
            "Y": [thickness],
            "model": [fixed_model, mixed_model],
            "contrast": [-age],
            "surf": [pial],
            "mask": [mask],
            "correction": [None, ["fdr", "rft"]],
            "niter": [1],
            "thetalim": [0.01],
            "drlim": [0.1],
            "two_tailed": [True],
            "cluster_threshold": [0.001, 1.2],
        },
    ]

    test_num = 0
    for params in ParameterGrid(parameters):
        test_num += 1
        slm = SLM(
            params["model"],
            params["contrast"],
            params["surf"],
            params["mask"],
            correction=params["correction"],
            niter=params["niter"],
            thetalim=params["thetalim"],
            drlim=params["drlim"],
            two_tailed=params["two_tailed"],
            cluster_threshold=params["cluster_threshold"],
        )
        slm.fit(params["Y"])

        # Save input/output
        if isinstance(params["model"], FixedEffect):
            params["model"] = age[:, None]
        else:
            params["model"] = np.concatenate((age[:, None], iq[:, None]), axis=1)
        dict2pkl(params, "slm", test_num, input=True)
        slm2files(slm, "slm", test_num, input=False)


if __name__ == "__main__":
    generate_test_data()



""" Tests for brainstat.datasets module. """
import numpy as np
import pytest
from brainspace.vtk_interface.wrappers.data_object import BSPolyData

from brainstat.datasets.base import (
    _valid_parcellations,
    fetch_parcellation,
    fetch_template_surface,
)

parametrize = pytest.mark.parametrize


@parametrize(
    "template",
    ["fslr32k", "fsaverage", "fsaverage3", "fsaverage4", "fsaverage5", "fsaverage6"],
)
def test_load_surfaces(template):
    """Test loading surface templates. For now this only tests whether it runs
    without error and whether the output type is correct."""
    surface = fetch_template_surface(template, join=True)
    assert isinstance(surface, BSPolyData)

    surface_lh, surface_rh = fetch_template_surface(template, join=False)

    assert isinstance(surface_lh, BSPolyData)
    assert isinstance(surface_rh, BSPolyData)


valid_n_regions = {k: v["n_regions"] for k, v in _valid_parcellations().items()}
valid_surfaces = {k: v["surfaces"] for k, v in _valid_parcellations().items()}


@parametrize("atlas", valid_n_regions.keys())
@parametrize("template", ["fsaverage", "fsaverage5", "fsaverage6", "fslr32k"])
def test_load_parcels(atlas, template):
    """Test loading surface parcels."""
    if template not in valid_surfaces[atlas]:
        pytest.skip(f"{atlas} does not exist on {template}")

    for n_regions in valid_n_regions[atlas]:
        parcels = fetch_parcellation(
            template, atlas, n_regions, join=True, seven_networks=True
        )
        assert isinstance(parcels, np.ndarray)

        parcels_lh, parcels_rh = fetch_parcellation(
            template, atlas, n_regions, join=False
        )
        assert isinstance(parcels_lh, np.ndarray)
        assert isinstance(parcels_rh, np.ndarray)

        assert parcels.size == parcels_lh.size + parcels_rh.size

        if atlas == "schaefer":
            parcels = fetch_parcellation(
                template, atlas, n_regions, join=True, seven_networks=False
            )
            assert isinstance(parcels, np.ndarray)



"""Classes for fixed, mixed, and random effects."""
import re
from typing import Any, List, Optional, Sequence, Union

import numpy as np
import pandas as pd

from brainstat._typing import ArrayLike
from brainstat._utils import deprecated


def check_names(
    x: Union[int, "FixedEffect", pd.DataFrame, pd.Series]
) -> Optional[List[str]]:
    """Return True if `x` is FixedEffect, Series or DataFrame."""
    if isinstance(x, (pd.DataFrame, pd.Series)):
        return x.columns.tolist()
    elif isinstance(x, FixedEffect):
        return x.names
    else:
        return None


def to_df(
    x: Union[int, ArrayLike, "FixedEffect"],
    n: int = 1,
    names: Optional[Union[str, Sequence[str]]] = None,
    idx: Optional[int] = None,
) -> pd.DataFrame:
    """Convert input to DataFrame.

    Parameters
    ----------
    x : int, array-like FixedEffect
        Input data.
    n : int, optional
        If input is a scalar, broadcast to column of `n` entries.
        Default is 1.
    names : str, sequence of str or None, optional
        Names for each column in `x`. Default is None.
    idx : int or None, optional
        Staring index for variable names of the for `x{i}`.

    Returns
    -------
    df : DataFrame
        Input `x` wrapped in a DataFrame.

    """

    if x is None:
        return pd.DataFrame()

    has_names = True
    if isinstance(x, pd.DataFrame):
        df = x
    elif isinstance(x, FixedEffect):
        df = x.m
    elif isinstance(x, pd.Series):
        df = pd.DataFrame(x)
    else:
        x_2d = np.atleast_2d(x)
        if x_2d.shape[0] == 1:
            x_2d = x_2d.T
        df = pd.DataFrame(x_2d)
        has_names = False

    if df.empty:
        return df

    if df.size == 1 and n > 1:
        df = pd.concat([df] * n, ignore_index=True)

    if names is None and (idx is not None or not has_names):
        if idx is None:
            idx = 0
        names = ["x%d" % i for i in range(idx, idx + df.shape[1])]

    if names is not None:
        if len(names) != df.shape[1]:
            raise ValueError(
                f"Number of columns {df.shape[1]} does not "
                f"coincide with column names {names}"
            )
        df.columns = names

    return pd.get_dummies(df)


def get_index(df: pd.DataFrame) -> Optional[int]:
    """Get index for column names of the form x{i}.

    If there are none, return 0.

    Parameters
    ----------
    df : DataFrame
        Input dataframe.

    Returns
    -------
    index : int
        Index for the next x column. If `df` is empty, return None.

    """
    if df.empty:
        return None
    r = [re.match(r"^x(\d+)$", c) for c in df.columns]
    r2 = [int(r1.groups()[0]) for r1 in r if r1 is not None]
    return 0 if len(r2) == 0 else max(r2) + 1


def check_duplicate_names(
    df1: pd.DataFrame, df2: Optional[pd.DataFrame] = None
) -> None:
    """Check columns with duplicate names.

    Parameters
    ----------
    df1 : DataFrame
        Input dataframe.
    df2 : DataFrame, optional
        If provided, check that dataframes do not contain columns with same
        names. Default is None.

    Raises
    ------
    ValueError
        If there are columns with duplicate names.

    """
    if df2 is None:
        names, counts = np.unique(df1.columns, return_counts=True)
        names = names[counts > 1]
    else:
        names = np.intersect1d(df1.columns, df2.columns)
    if names.size > 0:
        raise ValueError(f"Variables must have different names: {names}")


def remove_identical_columns(df1: pd.DataFrame, df2: pd.DataFrame) -> pd.DataFrame:
    """Remove columns with duplicate names across dataframes.

    Parameters
    ----------
    df1 : DataFrame
        Input dataframe from which to drop columns.
    df2 : DataFrame
        Reference dataFrame

    Returns
    -------
    DataFrame
        df1 with columns appearing in df2 removed.

    Raises
    ------
    ValueError
        If there are columns with duplicate names but no duplicate values.

    """
    names = np.intersect1d(df1.columns, df2.columns)
    for col in names:
        if np.array_equal(df1[col], df2[col]):
            df1 = df1.drop(col, axis=1)
        elif (df1[col].size == 1 or df2[col].size == 1) and np.all(
            df1[col].to_numpy() == df2[col].to_numpy()
        ):
            # Assume its an intercept term.
            df1 = df1.drop(col, axis=1)
        else:
            raise ValueError(
                f"Column {col} must be identical for duplicate removal. Either alter the column name or remove the duplicate data."
            )
    return df1


def remove_duplicate_columns(df: pd.DataFrame, tol: float = 1e-8) -> List[str]:
    """Remove duplicate columns.

    Parameters
    ----------
    df : DataFrame
        Input dataframe.
    tol : float, optional
        Tolerance to assess duplicate columns. Default is 1e-8.

    Returns
    -------
    columns: list of str
        Columns to keep after removing duplicates.

    """

    df = df / df.abs().sum(0)
    df *= 1 / tol
    # keep = df.round().T.drop_duplicates(keep="last").T.columns  # Slow!!
    idx = np.unique(df.round().values, axis=1, return_index=True)[-1]
    keep = df.columns[sorted(idx)]
    return keep


class FixedEffect(object):
    """Build a term object for a linear model.

    Parameters
    ----------
    x : array-like or DataFrame, optional
        If None, the term is empty. Default is None.
    names : str or list of str, optional
        Names for each column in `x`. If None, it defauts to {'x0', 'x1', ...}.
        Default is None.
    add_intercept : bool, optional
        If true, adds an intercept term. Defaults to True.

    Attributes
    ----------
    x : DataFrame
        Design matrix.
    names : list of str
        Names of columns in the design matrix.

    See Also
    --------
    MixedEffect: MixedEffect term

    Examples
    --------
    >>> t = FixedEffect()
    >>> t.is_empty
    True

    >>> t1 = FixedEffect(np.arange(5), names='t1')
    >>> t2 = FixedEffect(np.random.randn(5, 1), names=['t2'])
    >>> t3 = t1 + t2
    >>> t3.shape
    (5, 3)

    """

    tolerance = 1e-8

    def __init__(
        self,
        x: Optional[Union[ArrayLike, pd.DataFrame]] = None,
        names: Optional[Union[str, Sequence[str]]] = None,
        add_intercept: bool = True,
    ) -> None:

        if x is None:
            self.m = pd.DataFrame()
            return

        if isinstance(x, FixedEffect):
            self.m = x.m
            return

        if np.isscalar(x) and names is None:
            names = ["intercept"]

        if isinstance(names, str):
            names = [names]

        self.m = to_df(x, names=names).reset_index(drop=True)
        if add_intercept and "intercept" not in self.names:
            self.m.insert(0, "intercept", 1)
        check_duplicate_names(self.m)

    def _broadcast(
        self, t: Union[ArrayLike, "FixedEffect"], idx: Optional[int] = None
    ) -> pd.DataFrame:
        df = to_df(t, idx=idx)
        if self.shape[0] > 1 and df.shape[0] == 1:
            df = to_df(df, n=self.shape[0])
        elif self.shape[0] == 1 and df.shape[0] > 1:
            self.m = to_df(self.m, n=df.shape[0])
        elif not df.empty and self.shape[0] != df.shape[0]:
            raise ValueError(f"Cannot broadcast shape {df.shape} to " f"{self.shape}.")
        return df

    def _add(
        self, t: Union[ArrayLike, "FixedEffect", "MixedEffect"], side: str = "left"
    ) -> "FixedEffect":
        if isinstance(t, MixedEffect):
            return NotImplemented

        if self.empty:
            return FixedEffect(t, add_intercept=False)

        idx = None
        if check_names(t) is None:
            idx = get_index(self.m)

        df = self._broadcast(t, idx=idx)
        if df.empty:
            return self

        df = remove_identical_columns(df, self.m)

        # check_duplicate_names(df, df2=self.m)
        terms = [self.m, df]
        names = [self.names, list(df.columns)]
        if side == "right":
            terms = terms[::-1]
            names = names[::-1]
        df = pd.concat(terms, axis=1)
        df.columns = names[0] + names[1]
        cols = remove_duplicate_columns(df, tol=self.tolerance)
        return FixedEffect(df[cols], add_intercept=False)

    def __add__(
        self, t: Union[ArrayLike, "FixedEffect", "MixedEffect"]
    ) -> "FixedEffect":
        return self._add(t)

    def __radd__(
        self, t: Union[ArrayLike, "FixedEffect", "MixedEffect"]
    ) -> "FixedEffect":
        return self._add(t, side="right")

    def __sub__(self, t: Union[ArrayLike, "FixedEffect"]) -> "FixedEffect":
        if self.empty:
            return self
        df = self._broadcast(t)
        if df.empty:
            return self
        df /= df.abs().sum(0)
        df.index = self.m.index

        m = self.m / self.m.abs().sum(0)
        merged = m.T.merge(df.T, how="outer", indicator=True)
        mask = (merged._merge.values == "left_only")[: self.m.shape[1]]
        return FixedEffect(self.m[self.m.columns[mask]], add_intercept=False)

    def _mul(
        self, t: Union[ArrayLike, "FixedEffect", "MixedEffect"], side: str = "left"
    ) -> "FixedEffect":
        if isinstance(t, MixedEffect):
            return NotImplemented

        if self.is_empty:
            return self
        if np.isscalar(t):
            if t == 1:
                return self
            m = self.m * t
            if side == "right":
                names = [f"{t}*{k}" for k in self.names]
            else:
                names = [f"{k}*{t}" for k in self.names]
            return FixedEffect(m, names=names, add_intercept=False)

        df = self._broadcast(t)
        if df.empty:
            return FixedEffect()
        prod = []
        names = []
        for c in df.columns:
            prod.append(df[[c]].values * self.m)
            if side == "left":
                names.extend([f"{k}*{c}" for k in self.names])
            else:
                names.extend([f"{c}*{k}" for k in self.names])

        df = pd.concat(prod, axis=1)
        df.columns = names
        cols = remove_duplicate_columns(df, tol=self.tolerance)
        return FixedEffect(df[cols], add_intercept=False)

    def __mul__(
        self, t: Union[ArrayLike, "FixedEffect", "MixedEffect"]
    ) -> "FixedEffect":
        return self._mul(t)

    def __rmul__(
        self, t: Union[ArrayLike, "FixedEffect", "MixedEffect"]
    ) -> "FixedEffect":
        return self._mul(t, side="right")

    def __pow__(self, p: int) -> "FixedEffect":
        if p > 1:
            return self * self ** (p - 1)
        return self

    def __repr__(self) -> str:
        return self.m.__repr__()

    def _repr_html_(self) -> str:
        return self.m._repr_html_()

    @property
    def is_scalar(self) -> bool:
        return self.size == 1

    @property
    def matrix(self) -> pd.DataFrame:
        return self.m

    @property
    def names(self) -> List[str]:
        return object.__getattribute__(self, "m").columns.tolist()

    @property
    def is_empty(self) -> bool:
        return self.m.empty

    def __getattr__(self, name: str) -> Any:
        if name in object.__getattribute__(self, "names"):
            return object.__getattribute__(self, "m")[name].values
        if name in {"shape", "size", "empty"}:
            return getattr(self.m, name)
        return object.__getattribute__(self, name)


class MixedEffect:
    """Build a random term object for a linear model.

    Parameters
    ----------
    ran : array-like or DataFrame, optional
        For the random effects. If None, the random term is empty.
        Default is None.
    fix : array-like or DataFrame, optional
        If None, the fixed effects.
    name_ran : str, list, optional
        Name(s) for the random term(s). If None, it defauts to 'xi'.
        Default is None.
    name_fix : str, optional
        Name for the `fix` term. If None, it defauts to 'xi'.
        Default is None.
    ranisvar : bool, optional
        If True, `ran` is already a term for the variance. Default is False.

    Attributes
    ----------
    mean : FixedEffect
        FixedEffect for the mean.

    variance : FixedEffect
        FixedEffect for the variance.

    See Also
    --------
    FixedEffect: FixedEffect object

    Examples
    --------
    >>> r = MixedEffect()
    >>> r.is_empty
    True

    >>> r2 = MixedEffect(np.arange(5), name_ran='r1')
    >>> r2.mean.is_empty
    True
    >>> r2.variance.shape
    (25, 2)

    """

    def __init__(
        self,
        ran: Optional[Union[ArrayLike, pd.DataFrame]] = None,
        fix: Optional[Union[ArrayLike, pd.DataFrame]] = None,
        name_ran: Optional[Union[str, Sequence[str]]] = None,
        name_fix: Optional[Union[str, Sequence[str]]] = None,
        ranisvar: bool = False,
        add_intercept: bool = True,
        add_identity: bool = True,
    ) -> None:

        if isinstance(ran, MixedEffect):
            self.mean = ran.mean  # type: ignore
            self.variance = ran.variance  # type: ignore
            return

        if ran is None:
            self.variance = FixedEffect()
        else:
            ran = to_df(ran)
            if not ranisvar:
                if ran.size == 1:
                    name_ran = "I"
                    v = ran.values.flat[0]
                    if v != 1:
                        name_ran += f"{v}**2"
                else:
                    name = check_names(ran)
                    if name is not None and name_ran is None:
                        name_ran = name
                ran = ran @ ran.T
                ran = ran.values.ravel()

            self.variance = FixedEffect(ran, names=name_ran, add_intercept=False)
        self.mean = FixedEffect(fix, names=name_fix, add_intercept=add_intercept)

        if add_identity:
            I = MixedEffect(1, name_ran="I", add_identity=False)
            tmp_mixed = self + I
            self.variance = tmp_mixed.variance

        self.set_identity_last()

    def set_identity_last(self) -> None:
        """Sets the identity matrix column last.

        Raises
        ------
        ValueError
            Raised if "I" occurs more than once in the names.
        """

        if self.variance.is_empty:
            return

        if self.variance.m.size == 1:
            # Class is the identity.
            return

        n = int(round(np.sqrt(self.variance.m.shape[0])))
        I = np.reshape(np.identity(n), (-1, 1))
        index = np.argwhere(np.all(self.variance.m.to_numpy() == I, axis=0))

        if index.size == 0:
            return
        elif index.size == 1:
            names = self.variance.names
            names.append(names.pop(index[0][0]))
            self.variance.m = self.variance.m[names]
        else:
            raise ValueError("Found the identity matrix twice in the dataframe.")

    def broadcast_to(self, r1: "MixedEffect", r2: "MixedEffect") -> FixedEffect:
        if r1.variance.shape[0] == 1:
            v = np.eye(max(r2.shape[0], int(np.sqrt(r2.shape[2]))))
            return FixedEffect(v.ravel(), names="I", add_intercept=False)
        return r1.variance

    def _add(
        self, r: Union[FixedEffect, "MixedEffect"], side: str = "left"
    ) -> "MixedEffect":
        if not isinstance(r, MixedEffect):
            r = MixedEffect(fix=r, add_intercept=False, add_identity=False)

        r.variance = self.broadcast_to(r, self)
        self.variance = self.broadcast_to(self, r)
        if side == "left":
            ran = self.variance + r.variance
            fix = self.mean + r.mean
        else:
            ran = r.variance + self.variance
            fix = r.mean + self.mean

        s = MixedEffect(
            ran=ran, fix=fix, ranisvar=True, add_intercept=False, add_identity=False
        )
        s.set_identity_last()
        return s

    def __add__(self, r: Union[FixedEffect, "MixedEffect"]) -> "MixedEffect":  # type: ignore[misc]
        return self._add(r)

    def __radd__(self, r: Union[FixedEffect, "MixedEffect"]) -> "MixedEffect":  # type: ignore[misc]
        return self._add(r, side="right")

    def _sub(
        self, r: Union[FixedEffect, "MixedEffect"], side: str = "left"
    ) -> "MixedEffect":
        if not isinstance(r, MixedEffect):
            r = MixedEffect(fix=r, add_intercept=False, add_identity=False)
        r.variance = self.broadcast_to(r, self)
        self.variance = self.broadcast_to(self, r)
        if side == "left":
            ran = self.variance - r.variance
            fix = self.mean - r.mean
        else:
            ran = r.variance - self.variance
            fix = r.mean - self.mean
        return MixedEffect(
            ran=ran, fix=fix, ranisvar=True, add_intercept=False, add_identity=False
        )

    def __sub__(self, r: Union[FixedEffect, "MixedEffect"]) -> "MixedEffect":
        return self._sub(r)

    def __rsub__(self, r: Union[FixedEffect, "MixedEffect"]) -> "MixedEffect":
        return self._sub(r, side="right")

    def _mul(
        self, r: Union[FixedEffect, "MixedEffect"], side: str = "left"
    ) -> "MixedEffect":
        if not isinstance(r, MixedEffect):
            r = MixedEffect(fix=r, add_intercept=False, add_identity=False)
        r.variance = self.broadcast_to(r, self)
        self.variance = self.broadcast_to(self, r)

        if side == "left":
            ran = self.variance * r.variance
            fix = self.mean * r.mean
        else:
            ran = r.variance * self.variance
            fix = r.mean * self.mean
        s = MixedEffect(
            ran=ran, fix=fix, ranisvar=True, add_intercept=False, add_identity=False
        )

        if self.mean.matrix.values.size > 0:
            x = self.mean.matrix.values.T / self.mean.matrix.abs().values.max()
            t = FixedEffect()
            for i in range(x.shape[0]):
                for j in range(i + 1):
                    if i == j:
                        t = t + FixedEffect(
                            np.outer(x[i], x[j]).T.ravel(),
                            names=self.mean.names[i],
                            add_intercept=False,
                        )
                    else:
                        xs = x[i] + x[j]
                        xs_name = f"({self.mean.names[i]}+{self.mean.names[j]})"
                        xd = x[i] - x[j]
                        xd_name = f"({self.mean.names[i]}-{self.mean.names[j]})"

                        v = np.outer(xs, xs) / 4
                        t = t + FixedEffect(
                            v.ravel(), names=xs_name, add_intercept=False
                        )
                        v = np.outer(xd, xd) / 4
                        t = t + FixedEffect(
                            v.ravel(), names=xd_name, add_intercept=False
                        )

            s.variance = s.variance + t * r.variance

        if r.mean.matrix.values.size > 0:
            x = r.mean.matrix.values.T / r.mean.matrix.abs().values.max()
            t = FixedEffect()
            for i in range(x.shape[0]):
                for j in range(i + 1):
                    if i == j:
                        t = t + FixedEffect(
                            np.outer(x[i], x[j]).ravel(), names=r.mean.names[i]
                        )
                    else:
                        xs = x[i] + x[j]
                        xs_name = f"({r.mean.names[i]}+{r.mean.names[j]})"
                        xd = x[i] - x[j]
                        xd_name = f"({r.mean.names[i]}-{r.mean.names[j]})"

                        v = np.outer(xs, xs) / 4
                        t = t + FixedEffect(
                            v.ravel(), names=xs_name, add_intercept=False
                        )
                        v = np.outer(xd, xd) / 4
                        t = t + FixedEffect(
                            v.ravel(), names=xd_name, add_intercept=False
                        )
            s.variance = s.variance + self.variance * t
        s.set_identity_last()
        return s

    def __mul__(self, r: Union[FixedEffect, "MixedEffect"]) -> "MixedEffect":  # type: ignore[misc]
        return self._mul(r)

    def __rmul__(self, r: Union[FixedEffect, "MixedEffect"]) -> "MixedEffect":  # type: ignore[misc]
        return self._mul(r, side="right")

    def __pow__(self, p: int) -> "MixedEffect":
        if p > 1:
            return self * self ** (p - 1)
        return self

    @property
    def empty(self) -> bool:
        return self.mean.empty and self.variance.empty

    @property
    def shape(self) -> np.ndarray:
        return self.mean.shape + self.variance.shape

    def _repr_html_(self) -> str:
        return (
            "Mean:\n"
            + self.mean._repr_html_()
            + "\n\nVariance:\n"
            + self.variance._repr_html_()
        )


## Deprecated functions
@deprecated("Please use FixedEffect instead.")
def Term(x=None, names=None):
    return FixedEffect(x=x, names=names, add_intercept=False)


@deprecated("Please use MixedEffect instead.")
def Random(ran=None, fix=None, name_ran=None, name_fix=None, ranisvar=False):
    return MixedEffect(
        ran=ran,
        fix=fix,
        name_ran=name_ran,
        name_fix=name_fix,
        ranisvar=ranisvar,
        add_intercept=False,
        add_identity=False,
    )



""" Tests the fixed and mixed effects classes. """

import numpy as np

from brainstat.stats.terms import FixedEffect, MixedEffect


def test_fixed_init():
    """Tests the initialization of the FixedEffect class."""
    random_data = np.random.random_sample((10, 1))
    fix1 = FixedEffect(random_data, ["x0"])
    fix2 = FixedEffect(random_data, ["x0"], add_intercept=False)

    assert np.array_equal(fix1.m.shape, [10, 2])
    assert np.array_equal(fix1.names, ["intercept", "x0"])

    assert np.array_equal(fix2.m.shape, [10, 1])
    assert np.array_equal(fix2.names, ["x0"])


def test_fixed_overload():
    """Tests the overloads of the FixedEffect class."""
    random_data = np.random.random_sample((10, 3))
    fix01 = FixedEffect(random_data[:, :2], ["x0", "x1"], add_intercept=False)
    fix12 = FixedEffect(random_data[:, 1:], ["x2", "x3"], add_intercept=False)
    fix2 = FixedEffect(random_data[:, 2], ["x2"], add_intercept=False)
    fixi0 = FixedEffect(random_data[:, 0], ["x0"], add_intercept=True)
    fixi1 = FixedEffect(random_data[:, 1], ["x1"], add_intercept=True)

    fix_add = fix01 + fix12
    assert np.array_equal(fix_add.m, random_data)

    fix_add_intercept = 1 + FixedEffect(random_data[:, 0])
    assert np.array_equal(fixi0.m, fix_add_intercept.m)

    fix_add_intercept = fixi0 + fixi1
    expected = np.concatenate((np.ones((10, 1)), random_data[:, 0:2]), axis=1)
    assert np.array_equal(fix_add_intercept.m, expected)

    fix_sub = fix01 - fix12
    assert np.array_equal(fix_sub.m, random_data[:, 0][:, None])

    fix_mul = fix01 * fix2
    assert np.array_equal(fix_mul.m, random_data[:, :2] * random_data[:, 2][:, None])


def test_mixed_init():
    """Tests the initialization of the MixedEffect class."""
    n = 10
    random_data = np.random.random_sample((n, 1))
    mix1 = MixedEffect(random_data, ["x0"])
    mix2 = MixedEffect(random_data, ["x0"], add_identity=False)
    mix3 = MixedEffect(random_data, random_data, ["x0"], ["y0"])
    mix4 = MixedEffect(random_data, random_data, ["x0"], ["y0"], add_intercept=False)

    assert np.array_equal(mix1.variance.shape, [n ** 2, 2])
    assert np.array_equal(mix1.variance.names, ["x0", "I"])

    assert np.array_equal(mix2.variance.shape, [n ** 2, 1])
    assert np.array_equal(mix2.variance.names, ["x0"])

    assert np.array_equal(mix3.mean.shape, [10, 2])
    assert np.array_equal(mix3.mean.names, ["intercept", "y0"])

    assert np.array_equal(mix4.mean.shape, [10, 1])
    assert np.array_equal(mix4.mean.names, ["y0"])


def test_mixed_overload():
    """Tests the overloads of the MixedEffect class."""
    n = 3
    random_data = np.random.random_sample((n, 4))
    mix1 = MixedEffect(random_data[:, 0], name_ran=["x0"])
    mix2 = MixedEffect(random_data[:, 1], name_ran=["x1"])

    I = np.identity(n).flatten()[:, None]
    var12 = as_variance(random_data[:, :2])

    mix_add = mix1 + mix2
    expected_add = np.concatenate((var12, I), axis=1)
    assert np.array_equal(mix_add.variance.m, expected_add)

    mix_sub = mix1 - mix1
    assert mix_sub.empty

    mix_mul = mix1 * mix2
    expected_mul = np.concatenate(
        (
            var12[:, 0][:, None] * var12[:, 1][:, None],
            var12[:, 1][:, None] * I,
            var12[:, 0][:, None] * I,
            I,
        ),
        axis=1,
    )
    assert np.array_equal(mix_mul.variance.m, expected_mul)


def test_identity_detection():
    """Tests that the identity matrix is correctly placed last."""
    mix1 = MixedEffect(np.random.rand(3, 1), add_identity=False)
    mix2 = MixedEffect(1, name_ran="test_identity")
    I = np.identity(3).flatten()

    mix_add1 = mix2 + mix1
    mix_add2 = mix1 + mix2
    assert np.all(mix_add1.variance.m.to_numpy()[:, 1] == I)
    assert np.all(mix_add2.variance.m.to_numpy()[:, 1] == I)


def as_variance(M):
    var = [np.reshape(x[:, None] @ x[None, :], (-1, 1)) for x in M.T]
    return np.squeeze(var, axis=2).T



"""Utilities for running tests and test data generation."""
import pickle
from pathlib import Path

import numpy as np
import vtk
from brainspace.mesh.mesh_elements import get_cells, get_points
from brainspace.vtk_interface import wrap_vtk
from brainspace.vtk_interface.wrappers.data_object import BSPolyData

import brainstat
from brainstat.stats.SLM import SLM
from brainstat.stats.terms import FixedEffect, MixedEffect


def datadir(filename):
    """Returns the path to a given file in the test data directory.

    Parameters
    ----------
    filename : str
        Name of a file in the data directory.

    Returns
    -------
    str
        Full path to file in the data directory.
    """
    test_dir = Path(brainstat.__file__)
    repo_dir = test_dir.parents[1]
    return str(repo_dir / "extern" / "test-data" / filename)


def generate_slm(**kwargs):
    """Generates a SLM with the given attributes
    Parameters
    ----------
    All attributes of SLM can be provided as keyword arguments.
    Returns
    -------
    brainstat.stats.SLM.SLM
        SLM object.
    """
    slm = SLM(FixedEffect(1), 1)
    for key, value in kwargs.items():
        setattr(slm, key, value)
    return slm


def copy_slm(slm):
    """Copies an SLM object.
    Parameters
    ----------
    slm : brainstat.stats.SLM.SLM
        SLM object.
    Returns
    -------
    brainstat.stats.SLM.SLM
        SLM object.
    """
    slm_out = SLM(FixedEffect(1), 1)
    for key in slm.__dict__:
        setattr(slm_out, key, getattr(slm, key))
    return slm_out


def save_slm(slm, basename, testnum, input=True):
    """Saves an SLM object containing test data.
    Parameters
    ----------
    slm : brainstat.stats.SLM.SLM
        SLM object.
    basename : str
        Name for the tested function.
    testnum: int
        Test number.
    input: boolean, optional
        If True, appends _IN to filename. If false appends _OUT.
    Returns
    -------
    brainstat.stats.SLM.SLM
        SLM object.
    """
    D = slm2dict(slm)
    # Custom classes won't support MATLAB conversion.
    D.pop("surf", None)
    D.pop("_surf")
    if "_tri" in D:
        D["tri"] = D["_tri"]
        D.pop("_tri")
    D.pop("model", None)
    if input:
        stage = "IN"
    else:
        stage = "OUT"
    filename = datadir(basename + "_" + f"{testnum:02d}" + "_" + stage + ".pkl")
    with open(filename, "wb") as f:
        pickle.dump(D, f, protocol=4)


def slm2dict(slm):
    """Converts an SLM to a dictionary.
    Parameters
    ----------
    slm : brainstat.stats.SLM.SLM
        SLM object.
    Returns
    -------
    dict
        Dictionary with keys equivalent to the attributes of the slm.
    """
    D = {}
    for key in slm.__dict__:
        if getattr(slm, key) is not None:
            D[key] = getattr(slm, key)
    return D


def slm2files(slm, basename, test_num):
    """Converts an SLM to its output files.

    Parameters
    ----------
    slm : brainstat.stats.SLM
        SLM object.
    basename : str
        Base name for the file.
    test_num : int
        Number of the test.
    """
    D = slm2dict(slm)
    D.pop("model")
    D.pop("contrast")
    if "_surf" in D and isinstance(D["_surf"], BSPolyData):
        D["surf"] = {
            "tri": np.array(get_cells(D["_surf"])),
            "coord": np.array(get_points(D["_surf"])).T,
        }
        D.pop("_surf")
        D.pop("_tri")

    filename = datadir(basename + "_" + f"{test_num:02d}" + "_OUT.pkl")
    with open(filename, "wb") as f:
        pickle.dump(D, f, protocol=4)


def save_input_dict(params, basename, test_num):
    """Saves the input data.

    Parameters
    ----------
    params : dict
        Parameters provided by the parameter grid.
    basename : str
        Tag to save the file with.
    test_num : int
        Number of the test.
    """
    filename = datadir(basename + "_" + f"{test_num:02d}" + "_IN.pkl")

    if isinstance(params["surf"], BSPolyData):
        params["surf"] = {
            "tri": np.array(get_cells(params["surf"])) + 1,
            "coord": np.array(get_points(params["surf"])).T,
        }

    with open(filename, "wb") as f:
        pickle.dump(params, f, protocol=4)


def _generate_sphere():
    """Generates a vtk sphere of 50 vertices.

    Returns
    -------
    BSPolyData
        Mesh of a sphere.
    """
    s = vtk.vtkSphereSource()
    s.Update()
    return wrap_vtk(s.GetOutput())


def generate_random_data_model(
    n_observations,
    n_vertices,
    n_variates,
    n_predictors,
):
    """Generates random test data.

    Parameters
    ----------
    n_observations : int
        Number of observations.
    n_vertices : int
        Number of vertices.
    n_variates : int
        Number of variates.
    n_predictors : int
        Number of predictors.
    n_random : int
        Number of random effects.

    Returns
    -------
    numpy.ndarray
        Random data.
    numpy.ndarray
        Random model.
    """
    Y = np.random.random_sample((n_observations, n_vertices, n_variates))
    M = np.random.random_sample((n_observations, n_predictors))
    return Y, M


def array2effect(A, n_random=0):
    """Converts an input array to a set of effects.

    Parameters
    ----------
    A : np.array
        A samples-by-effects array.
    n_random : int, optional
        Number of random effects, by default 0. Random effects are selected from
        the first columns of A.

    Returns
    -------
    brainstat.stats.terms.FixedEffect, brainstat.stats.terms.MixedEffect
        The fixed/mixed effects.
    """
    fixed_effects = FixedEffect(A[:, n_random:])
    if n_random != 0:
        mixed_effects = MixedEffect(
            A[:, :n_random],
            name_ran=["f" + str(x) for x in range(n_random)],
        )
        return fixed_effects + mixed_effects
    else:
        return fixed_effects
